# -*- coding: utf-8 -*-
PLUGIN_NAME = "Tango.info Adapter"
PLUGIN_AUTHOR = "Felix Elsner, Sambhav Kothari, Philipp Wolfer"
PLUGIN_DESCRIPTION = """
<p>Load genre, date and vocalist tags for latin dance music
from <a href="https://tango.info">tango.info</a>.</p>
"""
PLUGIN_VERSION = "0.2.0"
PLUGIN_API_VERSIONS = ["2.6", "2.7"]

PLUGIN_LICENSE = "GPL-2.0"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.html"

import re
from functools import partial

from picard import log
from picard.util import LockableObject
from picard.metadata import register_track_metadata_processor


table_regex = re.compile(
       r'<h2><a href="\/tracks?">Tracks?<\/a><\/h2>(?!<\/table>)(.+?)<\/table>'
)  # Match the 'tracks'/'track' <table>
tr_regex = re.compile(r"<tr>((?!</tr>).+?)</tr>")  # Match <tr> elements
td_regex = re.compile(r"<td[^>]*>((?!</td>).+?)</td>")  # Match <td> elements
tint_regex = re.compile(r"TINT:([0-9]+-[0-9]{1,2}-[0-9]{1,2})")

TANGO_INFO_HOST = "tango.info"
TANGO_INFO_PORT = 443


# FIXME Remove all this
# This is all just a hack to get around server issues at tango.info
# The server will reply with an empty body if the Host header is set to
# "tango.info:443", but just "tango.info" is fine.
# Picard 2.8+ will by default not append the port to the Host header and thus
# not need this workaround
from picard import PICARD_VERSION
from picard.version import Version
host_workaround_needed = False
if PICARD_VERSION < Version(2, 8, 0, 'dev', 1):
    log.warning("%s: Picard version is older than 2.8, using workaround for "
                "tango.info Host header issue", PLUGIN_NAME)
    host_workaround_needed = True
if host_workaround_needed:
    from picard.webservice import WSGetRequest
    from PyQt5.QtCore import QUrl
    class WrappedWSGetRequest(WSGetRequest):  # noqa
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            url_without_port = QUrl(self.url().url())
            # -1 is QUrl defaultPort
            url_without_port.setPort(-1)
            self.setUrl(url_without_port)


class TangoInfoTagger:

    class TangoInfoScrapeQueue(LockableObject):

        def __init__(self):
            LockableObject.__init__(self)
            self.queue = {}

        def __contains__(self, name):
            return name in self.queue

        def __iter__(self):
            return self.queue.__iter__()

        def __getitem__(self, name):
            self.lock_for_read()
            value = self.queue.get(name)
            self.unlock()
            return value

        def __setitem__(self, name, value):
            self.lock_for_write()
            self.queue[name] = value
            self.unlock()

        def append(self, name, value):
            self.lock_for_write()
            if name in self.queue:
                self.queue[name].append(value)
                value = False
            else:
                self.queue[name] = [value]
                value = True
            self.unlock()
            return value

        def pop(self, name):
            self.lock_for_write()
            value = None
            if name in self.queue:
                value = self.queue[name]
                del self.queue[name]
            self.unlock()
            return value

    def __init__(self):
        self.albumpage_cache = {}
        self.albumpage_queue = self.TangoInfoScrapeQueue()

    def add_tangoinfo_data(self, album, track_metadata, track, release):

        # Look for BARCODE or barcode tag
        if track_metadata.get("barcode"):
            barcode = str(track_metadata["barcode"])
        elif track_metadata.get("BARCODE"):
            barcode = str(track_metadata["BARCODE"])
        else:
            # Abort if no barcode in track_metadata
            return

        # https://tango.info/wiki/TINT = <TINP>-<Side#>-<Track#>
        # https://tango.info/wiki/TINP - "Tango Info Number for a Product" is a
        # 14-digit numeric tango.info code used by tango.info and others.
        # Example: TINT:00743216335725-1-5
        # This plugin normalizes TINTs by stripping all zeros from the left
        tint = "%s-%s-%s" % (
            barcode.lstrip("0"),
            str(track_metadata.get("discnumber", "1")),
            str(track_metadata.get("tracknumber")),
        )

        if barcode in self.albumpage_cache:
            if self.albumpage_cache[barcode].get(tint):
                for field in ("genre", "date", "vocal"):
                    # Do no overwrite with empty data
                    if not self.albumpage_cache[barcode][tint].get(field):
                        continue
                    track_metadata[field] = \
                        self.albumpage_cache[barcode][tint][field]
            else:
                log.debug(
                    "%s: No information on tango.info for barcode %s",
                    PLUGIN_NAME, barcode)
        else:
            self.website_add_track(album, album._new_tracks[-1], barcode, tint)

    def website_add_track(self, album, track, barcode, tint, zeros=0):
        """
        :param zeros: Number of zeros that have been prepended to the barcode
        """
        self.album_add_request(album)

        if self.albumpage_queue.append(barcode, (track, album, tint)):

            # Barcode, zero-padded as needed
            path = "/%s" % ("0" * zeros) + barcode

            log.debug("%s: Downloading from tango.info: track %s, album %s "
                      "with TINT %s from %s",
                      PLUGIN_NAME, str(track), str(album), tint,
                      "https://%s%s" % (TANGO_INFO_HOST, path))
            # FIXME: Remove this
            if host_workaround_needed:
                ws = album.tagger.webservice
                def get_patched(self, *args, **kwargs):  # noqa
                    request = WrappedWSGetRequest(self, *args, **kwargs)
                    return ws.add_request(request)
                ws.get_patched = get_patched
                return ws.get_patched(
                    TANGO_INFO_HOST,
                    TANGO_INFO_PORT,
                    path,
                    partial(self.website_process, barcode, zeros),
                    priority=False,
                    important=False,
                    parse_response_type=None,
                    queryargs=None,
                )
            # Call website_process() as a partial func
            return album.tagger.webservice.get(
                TANGO_INFO_HOST,
                TANGO_INFO_PORT,
                path,
                partial(self.website_process, barcode, zeros),
                priority=False,
                important=False,
                parse_response_type=None,
                queryargs=None,
            )

    def website_process(self, barcode, zeros, response_bytes, reply, error):
        """
        response_bytes: PyQt5.QtCore.QByteArray, equals reply.readAll()
        reply:          PyQt5.QtNetwork.QNetworkReply
        error:          PyQt5.QtNetwork.QNetworkReply.NetworkError (optional)
        """

        if error:
            log.warning("%s: Network error retrieving info for barcode %s",
                        PLUGIN_NAME, barcode)
            track_triple = self.albumpage_queue.pop(barcode)
            for track, album, tint in track_triple:
                self.album_remove_request(album)
            return

        # Decode QByteArray into unicode
        response_decoded = response_bytes.data().decode('utf-8')

        tangoinfo_albumdata = self.extract_data(barcode, response_decoded)

        self.albumpage_cache[barcode] = tangoinfo_albumdata
        track_triple = self.albumpage_queue.pop(barcode)

        if tangoinfo_albumdata:
            if zeros:
                log.debug("%s: tango.info does not seem to have data for "
                          "barcode %s. "
                          "However, retrying with barcode %s (i.e. the same "
                          "with %s prepended) was successful. "
                          "This most likely means either MusicBrainz or "
                          "tango.info has stored a wrong barcode for this "
                          "release. You might want to investigate this "
                          "discrepancy and report it.",
                          PLUGIN_NAME, barcode, ("0" * zeros) + barcode,
                          ("a zero" if zeros == 1 else "%d zeros" % zeros))

            for track, album, tint in track_triple:
                tm = track.metadata
                if not self.albumpage_cache[barcode].get(tint):
                    self.album_remove_request(album)
                    continue

                for field in ("genre", "date", "vocal"):
                    # Write track metadata
                    if self.albumpage_cache[barcode][tint].get(field):
                        tm[field] = self.albumpage_cache[barcode][tint][field]

                for file in track.iterfiles():
                    fm = file.metadata
                    for field in ("genre", "date", "vocal"):
                        if not self.albumpage_cache[barcode][tint].get(field):
                            continue
                        # Write file metadata
                        fm[field] = self.albumpage_cache[barcode][tint][field]
                self.album_remove_request(album)
        else:
            if zeros >= 2:
                log.debug("%s: Could not load album with barcode %s even with "
                          "zero prepended(%s). This most likely means "
                          "tango.info does not have a release for this "
                          "barcode (or MusicBrainz has a wrong barcode)",
                          PLUGIN_NAME, barcode, ("0" * zeros) + barcode)
                for track, album, tint in track_triple:
                    self.album_remove_request(album)
                return

            log.debug("%s: Retrying with 0-padded barcode for barcode %s",
                      PLUGIN_NAME, barcode)

            for track, album, tint in track_triple:
                # Try again with zero-prepended barcode, but at most two times
                self.website_add_track(
                    album, track, barcode, tint, zeros=(zeros + 1)
                )
                self.album_remove_request(album)

    def album_add_request(self, album):
        album._requests += 1

    def album_remove_request(self, album):
        album._requests -= 1
        album._finalize_loading(None)

    def extract_data(self, barcode, response):

        # Check whether we have a concealed 404 and get the homepage
        if "<title>Contents - tango.info</title>" in response:
            log.debug("%s: No album with barcode %s on tango.info",
                      PLUGIN_NAME, barcode)
            return

        table = re.findall(table_regex, response)
        if not table:
            log.warning("%s: Could not extract table from album webpage - "
                        "regex failed or page structure changed", PLUGIN_NAME)
            return
        table = table[0]  # re.findall() returns a list

        # Content inside of <tr> elements
        trcontent = (match.groups()[0] for match in tr_regex.finditer(table))

        page_structure_warned = False  # Ratelimit warnings

        albuminfo = {}

        for tr in trcontent:
            # Content inside of <td> elements
            trackinfo = [m.groups()[0] for m in td_regex.finditer(tr)]

            # Example of expected structure:
            # <tr>
            # <td class="side_num">1</td>
            # <td class="track_num">6</td>
            # <td><a href="/T0370182390">Ese sos vos</a></td>
            # <td><a href="/genre.tango">tango</a></td>
            # <td><a href="/RicarTantu">Ricardo Tanturi</a></td>
            # <td><a href="/AlberDeluc">Alberto Castillo</a></td>
            # <td><a class="date" href="/1941-12-23">1941-12-23</a></td>
            # <td>02:38</td>
            # <td><a href="/00743216335725-1-6"
            #        title="TINT:00743216335725-1-6">info</a><br /></td>
            # </tr>

            # Sanity checks
            if not trackinfo:
                # Check if list is empty, e.g. contains a <th> for table header
                continue
            if len(trackinfo) < 9:
                if not page_structure_warned:  # Only warn once per <tr>
                    log.warning("%s: Table '<tr>' structure on webpage "
                                "unexpected for barcode %s",
                                PLUGIN_NAME, barcode)
                    page_structure_warned = True
                # Bail out early
                continue

            # Get tango.info TINT, e.g.
            # <a href="/00743216335725-1-4" title="TINT:00743216335725-1-4">
            tint = re.findall(tint_regex, trackinfo[8])
            if tint:
                # Normalize TINT by stripping all leading slashes and zeros
                tint = tint[0].lstrip("/").lstrip("0")
                albuminfo[tint] = {}
            else:
                # This really shouldn't happen
                log.warning("%s: No TINT found on webpage for barcode %s",
                            PLUGIN_NAME, barcode)
                continue

            # Get genre, e.g.
            # <a href="/genre.tango">tango</a>
            if trackinfo[3] != "-":
                genre = re.split("<|>", trackinfo[3])[2].title()
                albuminfo[tint]['genre'] = genre

            # Get date, e.g.
            # <a class="date" href="/1941-08-14">1941-08-14</a>
            if trackinfo[6] != "-":
                date = re.split("<|>", trackinfo[6])[2]
                albuminfo[tint]['date'] = date

            # Get singers, e.g.
            # <a href="/AlberDeluc">Alberto Castillo</a>
            if trackinfo[5] != "-":
                # Catch and strip <a> tags
                vocal = re.sub("<[^>]*>", "", trackinfo[5])
                albuminfo[tint]['vocal'] = vocal

        return albuminfo or None


tagger = TangoInfoTagger()
register_track_metadata_processor(tagger.add_tangoinfo_data)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'plugins/viewvariables/ui_variables_dialog.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_VariablesDialog(object):
    def setupUi(self, VariablesDialog):
        VariablesDialog.setObjectName("VariablesDialog")
        VariablesDialog.resize(600, 450)
        self.verticalLayout = QtWidgets.QVBoxLayout(VariablesDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.metadata_table = QtWidgets.QTableWidget(VariablesDialog)
        self.metadata_table.setAutoFillBackground(False)
        self.metadata_table.setSelectionMode(QtWidgets.QAbstractItemView.ContiguousSelection)
        self.metadata_table.setRowCount(1)
        self.metadata_table.setColumnCount(2)
        self.metadata_table.setObjectName("metadata_table")
        item = QtWidgets.QTableWidgetItem()
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        item.setFont(font)
        self.metadata_table.setHorizontalHeaderItem(0, item)
        item = QtWidgets.QTableWidgetItem()
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        item.setFont(font)
        self.metadata_table.setHorizontalHeaderItem(1, item)
        item = QtWidgets.QTableWidgetItem()
        item.setFlags(QtCore.Qt.ItemIsSelectable|QtCore.Qt.ItemIsEnabled)
        self.metadata_table.setItem(0, 0, item)
        item = QtWidgets.QTableWidgetItem()
        item.setFlags(QtCore.Qt.ItemIsSelectable|QtCore.Qt.ItemIsEnabled)
        self.metadata_table.setItem(0, 1, item)
        self.metadata_table.horizontalHeader().setDefaultSectionSize(150)
        self.metadata_table.horizontalHeader().setSortIndicatorShown(False)
        self.metadata_table.horizontalHeader().setStretchLastSection(True)
        self.metadata_table.verticalHeader().setVisible(False)
        self.metadata_table.verticalHeader().setDefaultSectionSize(20)
        self.metadata_table.verticalHeader().setMinimumSectionSize(20)
        self.verticalLayout.addWidget(self.metadata_table)
        self.buttonBox = QtWidgets.QDialogButtonBox(VariablesDialog)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(VariablesDialog)
        QtCore.QMetaObject.connectSlotsByName(VariablesDialog)

    def retranslateUi(self, VariablesDialog):
        _translate = QtCore.QCoreApplication.translate
        item = self.metadata_table.horizontalHeaderItem(0)
        item.setText(_translate("VariablesDialog", "Variable"))
        item = self.metadata_table.horizontalHeaderItem(1)
        item.setText(_translate("VariablesDialog", "Value"))
        __sortingEnabled = self.metadata_table.isSortingEnabled()
        self.metadata_table.setSortingEnabled(False)
        self.metadata_table.setSortingEnabled(__sortingEnabled)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'plugins/post_tagging_actions/options_post_tagging_actions.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtWidgets


class Ui_PostTaggingActions(object):
    def setupUi(self, PostTaggingActions):
        PostTaggingActions.setObjectName("PostTaggingActions")
        PostTaggingActions.resize(638, 450)
        PostTaggingActions.setMinimumSize(QtCore.QSize(100, 0))
        self.verticalLayout = QtWidgets.QVBoxLayout(PostTaggingActions)
        self.verticalLayout.setObjectName("verticalLayout")
        self.scrollArea = QtWidgets.QScrollArea(PostTaggingActions)
        self.scrollArea.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, -70, 606, 502))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.vboxlayout = QtWidgets.QVBoxLayout(self.scrollAreaWidgetContents)
        self.vboxlayout.setObjectName("vboxlayout")
        self.action_widget = QtWidgets.QWidget(self.scrollAreaWidgetContents)
        self.action_widget.setObjectName("action_widget")
        self._2 = QtWidgets.QVBoxLayout(self.action_widget)
        self._2.setObjectName("_2")
        self.widget_2 = QtWidgets.QWidget(self.action_widget)
        self.widget_2.setObjectName("widget_2")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout(self.widget_2)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.label_3 = QtWidgets.QLabel(self.widget_2)
        self.label_3.setObjectName("label_3")
        self.horizontalLayout_3.addWidget(self.label_3)
        self.add_file_path = QtWidgets.QPushButton(self.widget_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.add_file_path.sizePolicy().hasHeightForWidth())
        self.add_file_path.setSizePolicy(sizePolicy)
        self.add_file_path.setCheckable(False)
        self.add_file_path.setObjectName("add_file_path")
        self.horizontalLayout_3.addWidget(self.add_file_path)
        self._2.addWidget(self.widget_2)
        self.action = QtWidgets.QLineEdit(self.action_widget)
        self.action.setObjectName("action")
        self._2.addWidget(self.action)
        self.wait = QtWidgets.QCheckBox(self.action_widget)
        self.wait.setObjectName("wait")
        self._2.addWidget(self.wait)
        self.refresh = QtWidgets.QCheckBox(self.action_widget)
        self.refresh.setObjectName("refresh")
        self._2.addWidget(self.refresh)
        self.widget_3 = QtWidgets.QWidget(self.action_widget)
        self.widget_3.setObjectName("widget_3")
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout(self.widget_3)
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.albums = QtWidgets.QRadioButton(self.widget_3)
        self.albums.setObjectName("albums")
        self.horizontalLayout_6.addWidget(self.albums)
        self.tracks = QtWidgets.QRadioButton(self.widget_3)
        self.tracks.setObjectName("tracks")
        self.horizontalLayout_6.addWidget(self.tracks)
        self._2.addWidget(self.widget_3)
        self.vboxlayout.addWidget(self.action_widget)
        self.table_commands = QtWidgets.QWidget(self.scrollAreaWidgetContents)
        self.table_commands.setObjectName("table_commands")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.table_commands)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.widget = QtWidgets.QWidget(self.table_commands)
        self.widget.setObjectName("widget")
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout(self.widget)
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.add_action = QtWidgets.QPushButton(self.widget)
        self.add_action.setObjectName("add_action")
        self.horizontalLayout_5.addWidget(self.add_action)
        self.remove_action = QtWidgets.QPushButton(self.widget)
        self.remove_action.setObjectName("remove_action")
        self.horizontalLayout_5.addWidget(self.remove_action)
        self.horizontalLayout_2.addWidget(self.widget)
        self.widget1 = QtWidgets.QWidget(self.table_commands)
        self.widget1.setObjectName("widget1")
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout(self.widget1)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.up = QtWidgets.QToolButton(self.widget1)
        self.up.setText("")
        self.up.setArrowType(QtCore.Qt.UpArrow)
        self.up.setObjectName("up")
        self.horizontalLayout_4.addWidget(self.up)
        self.down = QtWidgets.QToolButton(self.widget1)
        self.down.setText("")
        self.down.setArrowType(QtCore.Qt.DownArrow)
        self.down.setObjectName("down")
        self.horizontalLayout_4.addWidget(self.down)
        self.horizontalLayout_2.addWidget(self.widget1, 0, QtCore.Qt.AlignRight)
        self.vboxlayout.addWidget(self.table_commands)
        self.table = QtWidgets.QTableWidget(self.scrollAreaWidgetContents)
        self.table.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContents)
        self.table.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.table.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.table.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.table.setObjectName("table")
        self.table.setColumnCount(4)
        self.table.setRowCount(0)
        item = QtWidgets.QTableWidgetItem()
        self.table.setHorizontalHeaderItem(0, item)
        item = QtWidgets.QTableWidgetItem()
        self.table.setHorizontalHeaderItem(1, item)
        item = QtWidgets.QTableWidgetItem()
        self.table.setHorizontalHeaderItem(2, item)
        item = QtWidgets.QTableWidgetItem()
        self.table.setHorizontalHeaderItem(3, item)
        self.table.horizontalHeader().setDefaultSectionSize(150)
        self.vboxlayout.addWidget(self.table)
        self.line = QtWidgets.QFrame(self.scrollAreaWidgetContents)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setObjectName("line")
        self.vboxlayout.addWidget(self.line)
        self.frame = QtWidgets.QFrame(self.scrollAreaWidgetContents)
        self.frame.setObjectName("frame")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.frame)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.cancel = QtWidgets.QCheckBox(self.frame)
        self.cancel.setObjectName("cancel")
        self.verticalLayout_2.addWidget(self.cancel)
        self.widget_4 = QtWidgets.QWidget(self.frame)
        self.widget_4.setObjectName("widget_4")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.widget_4)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.max_workers = QtWidgets.QSpinBox(self.widget_4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.max_workers.sizePolicy().hasHeightForWidth())
        self.max_workers.setSizePolicy(sizePolicy)
        self.max_workers.setMinimum(1)
        self.max_workers.setMaximum(64)
        self.max_workers.setObjectName("max_workers")
        self.horizontalLayout.addWidget(self.max_workers)
        self.label_2 = QtWidgets.QLabel(self.widget_4)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout.addWidget(self.label_2)
        self.verticalLayout_2.addWidget(self.widget_4)
        self.label = QtWidgets.QLabel(self.frame)
        self.label.setOpenExternalLinks(True)
        self.label.setObjectName("label")
        self.verticalLayout_2.addWidget(self.label)
        self.vboxlayout.addWidget(self.frame)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.vboxlayout.addItem(spacerItem)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.verticalLayout.addWidget(self.scrollArea)

        self.retranslateUi(PostTaggingActions)
        QtCore.QMetaObject.connectSlotsByName(PostTaggingActions)

    def retranslateUi(self, PostTaggingActions):
        _translate = QtCore.QCoreApplication.translate
        PostTaggingActions.setWindowTitle(_translate("PostTaggingActions", "Form"))
        self.label_3.setText(_translate("PostTaggingActions", "Insert action"))
        self.add_file_path.setToolTip(_translate("PostTaggingActions", "Add a file path to the command line."))
        self.add_file_path.setText(_translate("PostTaggingActions", "Add file"))
        self.wait.setToolTip(_translate("PostTaggingActions", "If checked, the next action runs immediately."))
        self.wait.setText(_translate("PostTaggingActions", " Wait for process to finish"))
        self.refresh.setToolTip(_translate("PostTaggingActions", "<html><head/><body><p>If checked, the album will &quot;refresh&quot; after this action finishes.</p></body></html>"))
        self.refresh.setText(_translate("PostTaggingActions", " Refresh tags after process finishes"))
        self.albums.setToolTip(_translate("PostTaggingActions", "Makes the action execute once for each album tagged."))
        self.albums.setText(_translate("PostTaggingActions", "Execute for albums"))
        self.tracks.setToolTip(_translate("PostTaggingActions", "Makes the action run once for each track tagged."))
        self.tracks.setText(_translate("PostTaggingActions", "Execute for tracks"))
        self.add_action.setToolTip(_translate("PostTaggingActions", "Add the action at the bottom of the queue."))
        self.add_action.setText(_translate("PostTaggingActions", "Add action"))
        self.remove_action.setToolTip(_translate("PostTaggingActions", "Remove the selected action."))
        self.remove_action.setText(_translate("PostTaggingActions", "Remove action"))
        self.table.setToolTip(_translate("PostTaggingActions", "Actions at the top of the list run first. Use the buttons on the right to reorder the selected action."))
        item = self.table.horizontalHeaderItem(0)
        item.setText(_translate("PostTaggingActions", "Action"))
        item = self.table.horizontalHeaderItem(1)
        item.setText(_translate("PostTaggingActions", "   Wait for exit   "))
        item = self.table.horizontalHeaderItem(2)
        item.setText(_translate("PostTaggingActions", "   Execute for tracks   "))
        item = self.table.horizontalHeaderItem(3)
        item.setText(_translate("PostTaggingActions", "   Refresh tags   "))
        self.cancel.setToolTip(_translate("PostTaggingActions", "<html><head/><body><p>If <span style=\" font-weight:700;\">not </span>checked, when Picard is closed, it will wait for the actions to finish in the background.</p></body></html>"))
        self.cancel.setText(_translate("PostTaggingActions", "Cancel actions in the queue when Picard is closed"))
        self.label_2.setToolTip(_translate("PostTaggingActions", "Sets the number of background threads executing the actions"))
        self.label_2.setText(_translate("PostTaggingActions", "   Maximum number of worker threads (Requires Picard restart)"))
        self.label.setText(_translate("PostTaggingActions", "<html><head/><body><p>Hover over each item to know more, or take a peek at the user guide <a href=\"https://github.com/metabrainz/picard-plugins/tree/2.0/plugins/post_tagging_actions/docs/guide.md\"><span style=\" text-decoration: underline; color:#3584e4;\">here.</span></a></p></body></html>"))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 Bob Swift (rdswift)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

# pylint: disable=missing-module-docstring, wrong-import-position, line-too-long

PLUGIN_NAME = 'Script Logger'
PLUGIN_AUTHOR = 'Bob Swift (rdswift)'
PLUGIN_DESCRIPTION = '''
This plugin provides a new script function `$logline()` to write entries
to Picard's system log.  By default, the log level is set at `Info`, but
any level can be used by providing the level as an optional second
parameter to the function.

The function is used as:

`$logline(text[,level])`

where `text` is the text to write to the log.  The entry will be written
at log level `Info` by default, but this can be changed by specifying a
different level as an optional second parameter.  Allowable log levels are:

- E (Error)
- W (Warning)
- I (Info)
- D (Debug)

If an unknown level is entered, the function will use the default `Info`
level.
'''
PLUGIN_VERSION = '0.1'
PLUGIN_API_VERSIONS = ['2.0', '2.1', '2.2', '2.3', '2.6', '2.9']
PLUGIN_LICENSE = "GPL-2.0"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.txt"

PLUGIN_USER_GUIDE_URL = "https://github.com/rdswift/picard-plugins/blob/2.0_RDS_Plugins/plugins/script_logger/docs/README.md"

DOCUMENTATION = """\
`$logline(text[,level])`

where `text` is the text to write to the log.  The entry will be written
at log level `Info` by default, but this can be changed by specifying a
different level as an optional second parameter.  Allowable log levels are:

- E (Error)
- W (Warning)
- I (Info)
- D (Debug)

If an unknown level is entered, the function will use the default `Info`
level.
"""

from picard import log
from picard.script import register_script_function

LEVELS = {
    'E': log.error,
    'W': log.warning,
    'I': log.info,
    'D': log.debug,
}

def logline(_parser, text: str, level=None):
    """Logs text to the Picard log.

    Args:
        _parser (parser): Script parser
        text (str): Text message to log
        level (str, optional): Level to use for logging ('E', 'W', 'I' or 'D'). Defaults to 'I'.
    """
    if level:
        _level = (str(level).strip().upper() + 'I')[0]
    else:
        _level = 'I'
    if _level not in LEVELS:
        _level = 'I'
    LEVELS[_level](text.strip())
    return ''

register_script_function(logline, documentation=DOCUMENTATION)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Bob Swift (rdswift)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

# pylint: disable=C0413     (wrong-import-position)
# pylint: disable=W0613     (unused-argument)

PLUGIN_NAME = 'Key Wheel Converter'
PLUGIN_AUTHOR = 'Bob Swift'
PLUGIN_DESCRIPTION = '''
Adds functions to convert between 'standard', 'camelot', 'open key' and 'traktor' key formats.
'''
PLUGIN_VERSION = '1.1'
PLUGIN_API_VERSIONS = ['2.3', '2.4', '2.6', '2.7']
PLUGIN_LICENSE = "GPL-2.0"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.txt"

import re

# pylint: disable=E0402     (import-error)
from picard import log
from picard.script import register_script_function


# pylint: disable=R0903     (too-few-public-methods)
class KeyMap():
    """
        Class to hold the mapping dictionary.  The dictionary is
        stored as a class variable so that it is only generated once.
    """

    # Circle of Fifths reference:
    # https://www.circleoffifths.com

    # Key Wheel references:
    # https://i.imgur.com/p9Kdevi.jpg
    # http://www.quanta.com.br/wp-content/uploads/2013/07/traktor-key-wheel_alta.jpg

    # List of tuples of:
    #   'camelot key',
    #   'open key',
    #   'standard key /w symbols',
    #   'standard key /w text'
    #   'traktor key'
    _keys = [
        ('1A', '6m', 'A♭ Minor', 'A-Flat Minor', 'Abm'),
        ('1B', '6d', 'B Major', 'B Major', 'B'),
        ('2A', '7m', 'E♭ Minor', 'E-Flat Minor', 'Ebm'),
        ('2B', '7d', 'F# Major', 'F-Sharp Major', 'F#'),
        ('3A', '8m', 'B♭ Minor', 'B-Flat Minor', 'Bbm'),
        ('3B', '8d', 'D♭ Major', 'D-Flat Major', 'Db'),
        ('4A', '9m', 'F Minor', 'F Minor', 'Fm'),
        ('4B', '9d', 'A♭ Major', 'A-Flat Major', 'Ab'),
        ('5A', '10m', 'C Minor', 'C Minor', 'Cm'),
        ('5B', '10d', 'E♭ Major', 'E-Flat Major', 'Eb'),
        ('6A', '11m', 'G Minor', 'G Minor', 'Gm'),
        ('6B', '11d', 'B♭ Major', 'B-Flat Major', 'Bb'),
        ('7A', '12m', 'D Minor', 'D Minor', 'Dm'),
        ('7B', '12d', 'F Major', 'F Major', 'F'),
        ('8A', '1m', 'A Minor', 'A Minor', 'Am'),
        ('8B', '1d', 'C Major', 'C Major', 'C'),
        ('9A', '2m', 'E Minor', 'E Minor', 'Em'),
        ('9B', '2d', 'G Major', 'G Major', 'G'),
        ('10A', '3m', 'B Minor', 'B Minor', 'Bm'),
        ('10B', '3d', 'D Major', 'D Major', 'D'),
        ('11A', '4m', 'G♭ Minor', 'G-Flat Minor', 'Gbm'),
        ('11B', '4d', 'A Major', 'A Major', 'A'),
        ('12A', '5m', 'D♭ Minor', 'D-Flat Minor', 'Dbm'),
        ('12B', '5d', 'E Major', 'E Major', 'E'),
    ]

    # Build mapping dictionary with 'camelot', 'standard with text'
    # and 'traktor' keys.
    keys = {}
    for item in _keys:
        for i in [0, 3, 4]:
            keys[item[i]] = {
                'camelot': item[0],
                'open': item[1],
                'standard_s': item[2],
                'standard_t': item[3],
                'traktor': item[4],
            }

    # Alternate mapping for standard keys
    s_alt = {
        'G-Flat Major': 'F-Sharp Major',
        'D-Sharp Minor': 'E-Flat Minor',
    }

    # Alternate mapping for traktor keys
    t_alt = {
        'G#': 'Ab',
        'A#': 'Bb',
        'C#': 'Db',
        'D#': 'Eb',
        'G#m': 'Abm',
        'A#m': 'Bbm',
        'C#m': 'Dbm',
        'D#m': 'Ebm',
        'F#m': 'Gbm',
    }


def _matcher(text, out_type):
    """Helper function that performs the actual key lookup.

    Args:
        text (str): Key provided by the user.
        out_type (str): Output format to use for the return value

    Returns:
        str: Value mapped to the key for the specified output type
    """
    match_text = _parse_input(text)
    if match_text not in KeyMap.keys:
        log.debug("{0}: Unable to match key: '{1}'".format(PLUGIN_NAME, text,))
        return ''
    return KeyMap.keys[match_text][out_type]


def _parse_input(text):
    """Helper function to parse the input argument to try to match
    one of the supported formats used for the mapping keys.

    Args:
        text (str): Input argument provided by the user

    Returns:
        str: Argument converted to supported key format (if possible)
    """

    text = text.strip()
    if not text:
        return ''

    if re.match("[0-9]{1,2}[ABab]$", text):
        # Matches camelot key.  Fix capitalization for lookup.
        return text.upper()

    if re.match("[0-9]{1,2}[dmDM]$", text):
        # Matches open key format.  Convert to camelot key for lookup.
        temp = int(text[0:-1])
        if 0 < temp < 13:
            _num = ((temp + 6) % 12) + 1
            _char = text[-1:].lower().replace('m', 'A').replace('d', 'B')
            return "{0}{1}".format(_num, _char,)

    if re.match("[a-g][#bB♭]?[mM]?$", text):
        # Matches Traktor key format.  Fix capitalization for lookup.
        temp = text[0:1].upper() + text[1:].replace('♭', 'b').lower()
        # Handle cases where there are multiple entries for the item
        if temp in KeyMap.t_alt:
            return KeyMap.t_alt[temp]
        return temp

    # Parse as standard key
    # Add missing hyphens before 'Flat' and 'Sharp'
    text = text.lower().replace(' s', '-s').replace(' f', '-f')
    # Convert symbols to text for lookup.
    parts = text.replace('♭', '-Flat').replace('#', '-Sharp').split()
    for (i, part) in enumerate(parts):
        parts[i] = part[0:1].upper() + part[1:]
    temp = ' '.join(parts).replace('-s', '-S').replace('-f', '-F')
    # Handle cases where there are multiple entries for the item
    if temp in KeyMap.s_alt:
        return KeyMap.s_alt[temp]
    return temp


def key2camelot(parser, text):
    """Any key to camelot format converter.

    Args:
        parser (object): Picard parser object
        text (str): Key to convert

    Returns:
        str: Converted key value

    Tests:

    >>> key2camelot(None, '1A')
    '1A'
    >>> key2camelot(None, '6m')
    '1A'
    >>> key2camelot(None, 'A♭ Minor')
    '1A'
    >>> key2camelot(None, 'A-Flat Minor')
    '1A'
    >>> key2camelot(None, '1a')
    '1A'
    >>> key2camelot(None, '6M')
    '1A'
    >>> key2camelot(None, 'A-FLAT MINOR')
    '1A'
    >>> key2camelot(None, 'a-flat minor')
    '1A'
    >>> key2camelot(None, ' 1A')
    '1A'
    >>> key2camelot(None, '1A ')
    '1A'
    >>> key2camelot(None, '')
    ''
    >>> key2camelot(None, 'A-Flat Minor x')
    ''
    >>> key2camelot(None, 'H-Flat Minor')
    ''
    >>> key2camelot(None, '1-Flat Minor')
    ''
    >>> key2camelot(None, 'A-Flat')
    ''
    >>> key2camelot(None, 'A♭')
    ''
    >>> key2camelot(None, '0a')
    ''
    >>> key2camelot(None, '13a')
    ''
    >>> key2camelot(None, '0m')
    ''
    >>> key2camelot(None, '13m')
    ''
    >>> key2camelot(None, '1x')
    ''

    >>> key2camelot(None, 'c')
    '8B'
    >>> key2camelot(None, 'dB')
    '3B'
    >>> key2camelot(None, 'd#M')
    '2A'
    """
    return _matcher(text, 'camelot')


def key2openkey(parser, text):
    """Any key to open key format converter.

    Args:
        parser (object): Picard parser object
        text (str): Key to convert

    Returns:
        str: Converted key value

    Tests:

    >>> key2openkey(None, '1A')
    '6m'
    >>> key2openkey(None, '6m')
    '6m'
    >>> key2openkey(None, 'A♭ Minor')
    '6m'
    >>> key2openkey(None, 'A-Flat Minor')
    '6m'
    >>> key2openkey(None, '1a')
    '6m'
    >>> key2openkey(None, '6M')
    '6m'
    >>> key2openkey(None, 'A-FLAT MINOR')
    '6m'
    >>> key2openkey(None, 'a-flat minor')
    '6m'
    >>> key2openkey(None, ' 1A')
    '6m'
    >>> key2openkey(None, '1A ')
    '6m'
    >>> key2openkey(None, '')
    ''
    >>> key2openkey(None, 'A-Flat Minor x')
    ''
    >>> key2openkey(None, 'H-Flat Minor')
    ''
    >>> key2openkey(None, '1-Flat Minor')
    ''
    >>> key2openkey(None, 'A-Flat')
    ''
    >>> key2openkey(None, 'A♭')
    ''
    >>> key2openkey(None, '0a')
    ''
    >>> key2openkey(None, '13a')
    ''
    >>> key2openkey(None, '0m')
    ''
    >>> key2openkey(None, '13m')
    ''
    >>> key2openkey(None, '1x')
    ''
    """
    return _matcher(text, 'open')


def key2standard(parser, text, use_symbol=''):
    """Any key to standard key format converter.

    Args:
        parser (object): Picard parser object
        text (str): Key to convert
        use_symbol (str, optional): Use '♭' and '#' symbols. Defaults to False.

    Returns:
        str: Converted key value

    Tests:

    >>> key2standard(None, '1A')
    'A-Flat Minor'
    >>> key2standard(None, '6m')
    'A-Flat Minor'
    >>> key2standard(None, 'A♭ Minor')
    'A-Flat Minor'
    >>> key2standard(None, 'A-Flat Minor')
    'A-Flat Minor'
    >>> key2standard(None, '1a')
    'A-Flat Minor'
    >>> key2standard(None, '6M')
    'A-Flat Minor'
    >>> key2standard(None, 'A-FLAT MINOR')
    'A-Flat Minor'
    >>> key2standard(None, 'a-flat minor')
    'A-Flat Minor'
    >>> key2standard(None, ' 1A')
    'A-Flat Minor'
    >>> key2standard(None, '1A ')
    'A-Flat Minor'
    >>> key2standard(None, '')
    ''
    >>> key2standard(None, 'A-Flat Minor x')
    ''
    >>> key2standard(None, 'H-Flat Minor')
    ''
    >>> key2standard(None, '1-Flat Minor')
    ''
    >>> key2standard(None, 'A-Flat')
    ''
    >>> key2standard(None, 'A♭')
    ''
    >>> key2standard(None, '0a')
    ''
    >>> key2standard(None, '13a')
    ''
    >>> key2standard(None, '0m')
    ''
    >>> key2standard(None, '13m')
    ''
    >>> key2standard(None, '1x')
    ''

    >>> key2standard(None, '1A', 'x')
    'A♭ Minor'
    >>> key2standard(None, '6m', 'x')
    'A♭ Minor'
    >>> key2standard(None, 'A♭ Minor', 'x')
    'A♭ Minor'
    >>> key2standard(None, 'A-Flat Minor', 'x')
    'A♭ Minor'
    >>> key2standard(None, '1a', 'x')
    'A♭ Minor'
    >>> key2standard(None, '6M', 'x')
    'A♭ Minor'
    >>> key2standard(None, 'A-FLAT MINOR', 'x')
    'A♭ Minor'
    >>> key2standard(None, 'a-flat minor', 'x')
    'A♭ Minor'
    >>> key2standard(None, ' 1A', 'x')
    'A♭ Minor'
    >>> key2standard(None, '1A ', 'x')
    'A♭ Minor'
    >>> key2standard(None, '1A ', ' ')
    'A♭ Minor'
    >>> key2standard(None, '', 'x')
    ''
    >>> key2standard(None, 'A-Flat Minor x', 'x')
    ''
    >>> key2standard(None, 'H-Flat Minor', 'x')
    ''
    >>> key2standard(None, '1-Flat Minor', 'x')
    ''
    >>> key2standard(None, 'A-Flat', 'x')
    ''
    >>> key2standard(None, 'A♭', 'x')
    ''
    >>> key2standard(None, '0a', 'x')
    ''
    >>> key2standard(None, '13a', 'x')
    ''
    >>> key2standard(None, '0m', 'x')
    ''
    >>> key2standard(None, '13m', 'x')
    ''
    >>> key2standard(None, '1x', 'x')
    ''

    >>> key2standard(None, 'A Flat Minor')
    'A-Flat Minor'
    >>> key2standard(None, 'F Sharp Major')
    'F-Sharp Major'
    >>> key2standard(None, 'G♭ Major')
    'F-Sharp Major'
    >>> key2standard(None, 'D# Minor')
    'E-Flat Minor'
    """
    if use_symbol:
        return _matcher(text, 'standard_s')
    return _matcher(text, 'standard_t')


def key2traktor(parser, text):
    """Any key to traktor key format converter.

    Args:
        parser (object): Picard parser object
        text (str): Key to convert

    Returns:
        str: Converted key value

    Tests:

    >>> key2traktor(None, '1A')
    'Abm'
    >>> key2traktor(None, '6m')
    'Abm'
    >>> key2traktor(None, 'A♭ Minor')
    'Abm'
    >>> key2traktor(None, 'A-Flat Minor')
    'Abm'
    >>> key2traktor(None, 'c#')
    'Db'
    >>> key2traktor(None, 'gBM')
    'Gbm'
    >>> key2traktor(None, 'g♭M')
    'Gbm'
    >>> key2traktor(None, '')
    ''
    """
    return _matcher(text, 'traktor')


register_script_function(key2camelot, name='key2camelot',
    documentation="""`$key2camelot(key)`

Returns the key string `key` in camelot key format.

The `key` argument can be entered in any of the supported formats, such as
'2B' (camelot), '6d' (open key), 'A♭ Minor' (standard with symbols),
'A-Flat Minor' (standard with text) or 'C#' (traktor).  If the `key` argument
is not recognized as one of the standard keys in the supported formats, then
an empty string will be returned.""")

register_script_function(key2openkey, name='key2openkey',
    documentation="""`$key2openkey(key)`

Returns the key string `key` in open key format.

The `key` argument can be entered in any of the supported formats, such as
'2B' (camelot), '6d' (open key), 'A♭ Minor' (standard with symbols),
'A-Flat Minor' (standard with text) or 'C#' (traktor).  If the `key` argument
is not recognized as one of the standard keys in the supported formats, then
an empty string will be returned.""")

register_script_function(key2standard, name='key2standard',
    documentation="""`$key2standard(key[,symbols])`

Returns the key string `key` in standard key format.  If the optional argument
`symbols` is set, then the '♭' and '#' symbols will be used, rather than
spelling out '-Flat' and '-Sharp'.

The `key` argument can be entered in any of the supported formats, such as
'2B' (camelot), '6d' (open key), 'A♭ Minor' (standard with symbols),
'A-Flat Minor' (standard with text) or 'C#' (traktor).  If the `key` argument
is not recognized as one of the standard keys in the supported formats, then
an empty string will be returned.""")

register_script_function(key2traktor, name='key2traktor',
    documentation="""`$key2traktor(key)`

Returns the key string `key` in traktor key format.

The `key` argument can be entered in any of the supported formats, such as
'2B' (camelot), '6d' (open key), 'A♭ Minor' (standard with symbols),
'A-Flat Minor' (standard with text) or 'C#' (traktor).  If the `key` argument
is not recognized as one of the standard keys in the supported formats, then
an empty string will be returned.""")

if __name__ == "__main__":
    import doctest
    doctest.testmod()



PLUGIN_NAME = 'AlbumArtist Extension'
PLUGIN_AUTHOR = 'Bob Swift (rdswift)'
PLUGIN_DESCRIPTION = '''
This plugin provides standardized, credited and sorted artist information
for the album artist.  This is useful when your tagging or renaming scripts
require both the standardized artist name and the credited artist name, or
more detailed information about the album artists.
<br /><br />
The information is provided in the following variables:
<ul>
<li>_aaeStdAlbumArtists = The standardized version of the album artists.
<li>_aaeCredAlbumArtists = The credited version of the album artists.
<li>_aaeSortAlbumArtists = The sorted version of the album artists.
<li>_aaeStdPrimaryAlbumArtist = The standardized version of the first
    (primary) album artist.
<li>_aaeCredPrimaryAlbumArtist = The credited version of the first (primary)
    album artist.
<li>_aaeSortPrimaryAlbumArtist = The sorted version of the first (primary)
    album artist.
<li>_aaeAlbumArtistCount = The number of artists comprising the album artist.
</ul>
PLEASE NOTE: Once the plugin is installed, it automatically makes these 
variables available to File Naming Scripts and other scripts in Picard. 
Like other variables, you must mention them in a script for them to affect 
the file name or other data.
<br /><br />
This plugin is no longer being maintained. 
Consider switching to the 
[Additional Artists Variables plugin](https://github.com/rdswift/picard-plugins/tree/2.0/plugins/additional_artists_variables), 
which fills this 
role, and also includes additional variables. That other plugin uses different 
names for the album artist names provided here, so you if you switch plugins, you
will need to update your scripts with the different names.
<br /><br />
Version 0.6.1 of this plugin functions identically to Version 0.6. Only this 
description (and the version number) has changed.
'''

PLUGIN_VERSION = "0.6.1"
PLUGIN_API_VERSIONS = ["2.0"]
PLUGIN_LICENSE = "GPL-2.0-or-later"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.html"

from picard import config, log
from picard.metadata import register_album_metadata_processor
from picard.plugin import PluginPriority

def add_artist_std_name(album, album_metadata, release_metadata):
    album_id = release_metadata['id']
    # Test for valid metadata node for the release
    if 'artist-credit' in release_metadata:
        # Initialize variables to default values
        cred_artist = ""
        std_artist = ""
        sort_artist = ""
        artist_count = 0
        # The 'artist-credit' key should always be there.
        # This check is to avoid a runtime error if it doesn't exist for some reason.
        for artist_credit in release_metadata['artist-credit']:
            # Initialize temporary variables for each loop.
            temp_std_name = ""
            temp_cred_name = ""
            temp_sort_name = ""
            temp_phrase = ""
            # Check if there is a 'joinphrase' specified.
            if 'joinphrase' in artist_credit:
                temp_phrase = artist_credit['joinphrase']
            else:
                metadata_error(album_id, 'artist-credit.joinphrase')
            # Check if there is a 'name' specified.
            if 'name' in artist_credit:
                temp_cred_name = artist_credit['name']
            else:
                metadata_error(album_id, 'artist-credit.name')
            # Check if there is an 'artist' specified.
            if 'artist' in artist_credit:
                # Check if there is a 'name' specified.
                if 'name' in artist_credit['artist']:
                    temp_std_name = artist_credit['artist']['name']
                else:
                    metadata_error(album_id, 'artist-credit.artist.name')
                if 'sort-name' in artist_credit['artist']:
                    temp_sort_name = artist_credit['artist']['sort-name']
                else:
                    metadata_error(album_id, 'artist-credit.artist.sort-name')
            else:
                metadata_error(album_id, 'artist-credit.artist')
            std_artist += temp_std_name + temp_phrase
            cred_artist += temp_cred_name + temp_phrase
            sort_artist += temp_sort_name + temp_phrase
            if artist_count < 1:
                album_metadata["~aaeStdPrimaryAlbumArtist"] = temp_std_name
                album_metadata["~aaeCredPrimaryAlbumArtist"] = temp_cred_name
                album_metadata["~aaeSortPrimaryAlbumArtist"] = temp_sort_name
            artist_count += 1
    else:
        metadata_error(album_id, 'artist-credit')
    if std_artist:
        album_metadata["~aaeStdAlbumArtists"] = std_artist
    if cred_artist:
        album_metadata["~aaeCredAlbumArtists"] = cred_artist
    if sort_artist:
        album_metadata["~aaeSortAlbumArtists"] = sort_artist
    if artist_count:
        album_metadata["~aaeAlbumArtistCount"] = artist_count
    return None


def metadata_error(album_id, metadata_element):
    log.error("%s: %r: Missing '%s' in release metadata.",
            PLUGIN_NAME, album_id, metadata_element)

# Register the plugin to run at a LOW priority so that other plugins that
# modify the artist information can complete their processing and this plugin
# is working with the latest updated data.
register_album_metadata_processor(add_artist_std_name, priority=PluginPriority.LOW)



# -*- coding: utf-8 -*-

PLUGIN_NAME = "Add Cluster As Release"
PLUGIN_AUTHOR = 'Frederik "Freso" S. Olesen, Lukáš Lalinský, Philip Jägenstedt'
PLUGIN_DESCRIPTION = "Adds a plugin context menu option to clusters and single\
 files to help you quickly add them as releases or standalone recordings to\
 the MusicBrainz database via the website by pre-populating artists,\
 track names and times."
PLUGIN_VERSION = "0.7.3"
PLUGIN_API_VERSIONS = ["2.0"]

from picard import config, log
from picard.cluster import Cluster
from picard.const import MUSICBRAINZ_SERVERS
from picard.file import File
from picard.util import webbrowser2
from picard.ui.itemviews import BaseAction, register_cluster_action, register_file_action

import os
import tempfile

HTML_HEAD = """<!doctype html>
<meta charset="UTF-8">
<title>%s</title>
<form action="%s" method="post">
"""
HTML_INPUT = """<input type="hidden" name="%s" value="%s">
"""
HTML_TAIL = """<input type="submit" value="%s">
</form>
<script>document.forms[0].submit()</script>
"""
HTML_ATTR_ESCAPE = {
    "&": "&amp;",
    '"': "&quot;"
}


def mbserver_url(path):
    host = config.setting["server_host"]
    port = config.setting["server_port"]
    if host in MUSICBRAINZ_SERVERS or port == 443:
        urlstring = "https://%s%s" % (host, path)
    elif port is None or port == 80:
        urlstring = "http://%s%s" % (host, path)
    else:
        urlstring = "http://%s:%d%s" % (host, port, path)
    return urlstring


class AddObjectAsEntity(BaseAction):
    NAME = "Add Object As Entity..."
    objtype = None
    submit_path = '/'

    def __init__(self):
        super(AddObjectAsEntity, self).__init__()
        self.form_values = {}

    def check_object(self, objs, objtype):
        """
        Checks if a given object array is valid (ie., has one item) and that
        its item is an object of the given type.

        Returns either False (if conditions are not met), or the object in the
        array.
        """
        if not isinstance(objs[0], objtype) or len(objs) != 1:
            return False
        else:
            return objs[0]

    def add_form_value(self, key, value):
        "Add global (e.g., release level) name-value pair."
        self.form_values[key] = value

    def set_form_values(self, objdata):
        return

    def generate_html_file(self, form_values):
        (fd, fp) = tempfile.mkstemp(suffix=".html")

        with os.fdopen(fd, "w", encoding="utf-8") as f:
            def esc(s):
                return "".join(HTML_ATTR_ESCAPE.get(c, c) for c in s)
            # add a global (release-level) name-value

            def nv(n, v):
                f.write(HTML_INPUT % (esc(n), esc(v)))

            f.write(HTML_HEAD % (self.NAME, mbserver_url(self.submit_path)))

            for key in form_values:
                nv(key, form_values[key])

            f.write(HTML_TAIL % (self.NAME))

        return fp

    def open_html_file(self, fp):
        webbrowser2.open("file://" + fp)

    def callback(self, objs):
        objdata = self.check_object(objs, self.objtype)
        try:
            if objdata:
                self.set_form_values(objdata)
                html_file = self.generate_html_file(self.form_values)
                self.open_html_file(html_file)
        finally:
            self.form_values.clear()


class AddClusterAsRelease(AddObjectAsEntity):
    NAME = "Add Cluster As Release..."
    objtype = Cluster
    submit_path = '/release/add'

    def __init__(self):
        super().__init__()
        self.discnumber_shift = -1

    def extract_discnumber(self, metadata):
        """
        >>> from picard.metadata import Metadata
        >>> m = Metadata()
        >>> AddClusterAsRelease().extract_discnumber(m)
        0
        >>> m["discnumber"] = "boop"
        >>> AddClusterAsRelease().extract_discnumber(m)
        0
        >>> m["discnumber"] = "1"
        >>> AddClusterAsRelease().extract_discnumber(m)
        0
        >>> m["discnumber"] = 1
        >>> AddClusterAsRelease().extract_discnumber(m)
        0
        >>> m["discnumber"] = -1
        >>> AddClusterAsRelease().extract_discnumber(m)
        0
        >>> m["discnumber"] = "1/1"
        >>> AddClusterAsRelease().extract_discnumber(m)
        0
        >>> m["discnumber"] = "2/2"
        >>> AddClusterAsRelease().extract_discnumber(m)
        1
        >>> a = AddClusterAsRelease()
        >>> m["discnumber"] = "-2/2"
        >>> a.extract_discnumber(m)
        0
        >>> m["discnumber"] = "-1/4"
        >>> a.extract_discnumber(m)
        1
        >>> m["discnumber"] = "1/4"
        >>> a.extract_discnumber(m)
        3

        """
        # As per https://musicbrainz.org/doc/Development/Release_Editor_Seeding#Tracklists_data
        # the medium numbers ("m") must be starting with 0.
        # Maybe the existing tags don't have disc numbers in them or
        # they're starting with something smaller than or equal to 0, so try
        # to produce a sane disc number.
        try:
            discnumber = metadata.get("discnumber", "1")
            # Split off any totaldiscs information
            discnumber = discnumber.split("/", 1)[0]
            m = int(discnumber)
            if m <= 0:
                # A disc number was smaller than or equal to 0 - all other
                # disc numbers need to be changed to accommodate that.
                self.discnumber_shift = max(self.discnumber_shift, 0 - m)
            m = m + self.discnumber_shift
        except ValueError as e:
            # The most likely reason for an exception at this point is because
            # the disc number in the tags was not a number. Just log the
            # exception and assume the medium number is 0.
            log.info("Trying to get the disc number of %s caused the following error: %s; assuming 0",
                     metadata["~filename"], e)
            m = 0
        return m

    def set_form_values(self, cluster):
        nv = self.add_form_value

        nv("artist_credit.names.0.artist.name", cluster.metadata["albumartist"])
        nv("name", cluster.metadata["album"])

        for i, file in enumerate(cluster.files):
            try:
                i = int(file.metadata["tracknumber"]) - 1
            except:
                pass

            m = self.extract_discnumber(file.metadata)

            # add a track-level name-value
            def tnv(n, v):
                nv("mediums.%d.track.%d.%s" % (m, i, n), v)

            tnv("name", file.metadata["title"])
            if file.metadata["artist"] != cluster.metadata["albumartist"]:
                tnv("artist_credit.names.0.name", file.metadata["artist"])
            tnv("length", str(file.metadata.length))


class AddFileAsRecording(AddObjectAsEntity):
    NAME = "Add File As Standalone Recording..."
    objtype = File
    submit_path = '/recording/create'

    def set_form_values(self, track):
        nv = self.add_form_value
        nv("edit-recording.name", track.metadata["title"])
        nv("edit-recording.artist_credit.names.0.artist.name", track.metadata["artist"])
        nv("edit-recording.length", track.metadata["~length"])


class AddFileAsRelease(AddObjectAsEntity):
    NAME = "Add File As Release..."
    objtype = File
    submit_path = '/release/add'

    def set_form_values(self, track):
        nv = self.add_form_value

        # Main album attributes
        if track.metadata["albumartist"]:
            nv("artist_credit.names.0.artist.name", track.metadata["albumartist"])
        else:
            nv("artist_credit.names.0.artist.name", track.metadata["artist"])
        if track.metadata["album"]:
            nv("name", track.metadata["album"])
        else:
            nv("name", track.metadata["title"])

        # Tracklist
        nv("mediums.0.track.0.name", track.metadata["title"])
        nv("mediums.0.track.0.artist_credit.names.0.name", track.metadata["artist"])
        nv("mediums.0.track.0.length", str(track.metadata.length))


register_cluster_action(AddClusterAsRelease())
register_file_action(AddFileAsRecording())
register_file_action(AddFileAsRelease())

if __name__ == "__main__":
    import doctest
    doctest.testmod()



from contextlib import contextmanager
from typing import Generator


@contextmanager
def override_module(obj: object) -> Generator[None, None, None]:
    # picard expects hooks to be defined at module level
    module = obj.__module__
    obj.__module__ = ".".join(module.split(".")[:-1])
    yield
    obj.__module__ = module



PLUGIN_NAME = 'Feat. Artists Removed'
PLUGIN_AUTHOR = 'Lukas Lalinsky, Bryan Toth'
PLUGIN_DESCRIPTION = 'Removes feat. artists from track titles.  Substitution is case insensitive.'
PLUGIN_VERSION = "0.4"
PLUGIN_API_VERSIONS = ["0.9.0", "0.10", "0.15", "0.16", "2.0"]

from picard.metadata import register_track_metadata_processor
import re

_feat_re = re.compile(r"\s+\(feat\. [^)]*\)", re.IGNORECASE)


def remove_featartists(tagger, metadata, track, release):
    metadata["title"] = _feat_re.sub("", metadata["title"])

register_track_metadata_processor(remove_featartists)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Giorgio Fontanive (twodoorcoupe)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

PLUGIN_NAME = "Enhanced Titles"
PLUGIN_AUTHOR = "Giorgio Fontanive"
PLUGIN_DESCRIPTION = """
This plugin sets the albumsort and titlesort tags. It also provides the script
functions $swapprefix_lang, $delprefix_lang and $title_lang. The languages
included at the moment are English, Spanish, Italian, German, French and Portuguese.

The functions do the same thing as their original counterparts, but take
multiple languages as parameters. If no languages are provided, all the available ones are
included. Languages are provided with ISO 639-3 codes: eng, spa, ita, fra, deu, por.

Tagging and checking aliases can be disabled in the plugin's options page, found
under "plugins". Checking aliases will slow down processing.

If you wish to add your own language or to change the words that are not capitalized,
please feel free to code the changes and submit a pull request along with links to web
pages that give definitive language-specific title-case rules.
"""
PLUGIN_VERSION = "0.1"
PLUGIN_API_VERSIONS = ["2.10"]
PLUGIN_LICENSE = "GPL-2.0"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.html"

from picard import log, config
from picard.plugin import PluginPriority
from picard.metadata import register_album_metadata_processor, register_track_metadata_processor
from picard.script import register_script_function
from picard.script.functions import func_swapprefix, func_delprefix
from picard.webservice.api_helpers import MBAPIHelper

from picard.ui.options import OptionsPage, register_options_page
from .ui_options_enhanced_titles import Ui_EnhancedTitlesOptions

from functools import partial
import re

# Options.
KEEP_ALLCAPS = "et_keep_allcaps"
ENABLE_TAGGING = "et_enable_tagging"
CHECK_ALBUM = "et_check_album_aliases"
CHECK_TRACK = "et_check_track_aliases"

_articles = {
    "eng": {"the", "a", "an"},
    "spa": {"el", "los", "la", "las", "lo", "un", "unos", "una", "unas"},
    "ita": {"il", "l'", "la", "i", "gli", "le", "un", "uno", "una", "un'"},
    "fra": {"le", "la", "les", "un", "une", "des", "l'"},
    "deu": {"der", "den", "die", "das", "dem", "des", "den"},
    "por": {"o", "os", "a", "as", "um", "uns", "uma", "umas"}
}

# Prepositions and conjunctions with 3 letters or fewer.
# These will stay in lower case when doing title case.
_other_minor_words = {
    "eng": {"so", "yet", "as", "at", "by", "for", "in", "of", "off", "on", "per",
            "to", "up", "via", "and", "as", "but", "for", "if", "nor", "or", "en", "via", "vs"},
    "spa": {"mas", "que", "en", "con", "por", "de", "y", "e", "o", "u", "si", "ni"},
    "ita": {"di", "a", "da", "in", "con", "su", "per", "tra", "fra", "e", "o", "ma", "se"},
    "fra": {"à", "de", "en", "par", "sur", "et", "ou", "que", "si"},
    "deu": {"bis", "für", "um", "an", "auf", "in", "vor", "aus", "bei", "mit",
            "von", "zu", "la", "so", "daß", "als", "ob", "ehe"},
    "por": {"dem", "em", "por", "ao", "à", "aos", "às", "do", "da", "dos", "das",
            "no", "na", "nos", "nas", "num", "dum", "e", "mas", "até", "em", "ou",
            "que", "se", "por"}
}


def flatten_values(dictionary):
    l = list()
    for v in dictionary.values():
        l += v
    return set(l)


_articles_langs = set(_articles)
_all_articles = flatten_values(_articles)
_all_other_minor_words = flatten_values(_other_minor_words)


class ReleaseGroupHelper(MBAPIHelper):
    """API Helper to retrieve release group information.
    """

    def get_release_group_by_id(self, release_id, handler, inc = None):
        """Gets the information for a release group.
        """
        return self._get_by_id("release-group", release_id, handler, inc)


class SortTagger:
    """Sets the titlesort and albumsort tags.

    First, it checks if there is already a sort name available in one of the
    aliases. If it does not find any, it swaps the prefix if there is one in
    the title.
    """

    def _select_alias(self, aliases, name):
        """Selects the first alias where the names match and the sort name is
        different.

        Args:
            aliases (list): One dictionary for each alias available.
            name (str): Title of the album/track.

        Returns:
            (str): The sort name of the first useful alias it finds. None if
                   none are found.

        For example, the album "The Beatles" has alias "Le Double Blanc" with
        sort name "Double Blanc, Le", so it's not considered. But it also has alias
        "The Beatles" with sort name "Beatles, The", so this one is chosen.
        Another example, "The Continuing Story of Bungalow Bill" has an alias
        with the sort name equal to the title, this is not considered because
        it makes more sense to swap the prefix.
        """
        name_casefold = name.casefold()
        for alias in aliases:
            sortname = alias["sort-name"]
            if (alias["name"].casefold() == name_casefold and
                    not sortname.casefold() == name_casefold):
                log.info('Enhanced Titles: sort name found for "%s", "%s".', name, sortname)
                return sortname
        log.info('Enhanced Titles: no proper sort name found for "%s".', name)
        return None

    def _response_handler(self, document, reply, error, metadata = None, field = None):
        """Handles the response from MusicBrainz.

        Args:
            metadata (MetaData): The object that needs to be updated.
            field (str): Either "title" or "album", depending on what is being
                         updated.
        """
        sortname = None
        try:
            if document:
                if error:
                    log.error("Enhanced Titles: information retrieval error.")
                if document["aliases"]:
                    sortname = self._select_alias(document["aliases"], metadata[field])
                else:
                    log.info('Enhanced Titles: no aliases found for "%s".', metadata[field])
        finally:
            sortfield = field + "sort"
            if sortname:
                metadata[sortfield] = sortname
            else:
                metadata[sortfield] = self._swapprefix(metadata, field)

    def _swapprefix(self, metadata, field):
        """Swaps the prefix of the title based on the album/track language.

        If none of the languages are available, it just copies the title.
        Otherwise, if no language information is found, then it uses all languages available.
        Otherwise, it uses exclusively the languages that are also available.
        """
        languages = lang_functions.find_languages(metadata)
        if not languages:  # None of the languages found are available.
            return metadata[field]
        prefixes = lang_functions.find_prefixes(languages)
        return func_swapprefix(None, metadata[field], *prefixes)

    def set_track_titlesort(self, album, metadata, track, release):
        """Sets the track's titlesort field.
        """
        if config.setting[ENABLE_TAGGING]:
            handler = partial(self._response_handler, metadata = metadata, field = "title")
            if config.setting[CHECK_TRACK]:
                MBAPIHelper(album.tagger.webservice).get_track_by_id(
                    metadata["musicbrainz_recordingid"],
                    handler,
                    inc = ["aliases"]
                )
            else:
                metadata["titlesort"] = self._swapprefix(metadata, "title")

    def set_album_titlesort(self, album, metadata, release):
        """Sets the album's albumsort field.
        """
        if config.setting[ENABLE_TAGGING]:
            handler = partial(self._response_handler, metadata = metadata, field = "album")
            if config.setting[CHECK_ALBUM]:
                ReleaseGroupHelper(album.tagger.webservice).get_release_group_by_id(
                    metadata["musicbrainz_releasegroupid"],
                    handler,
                    inc = ["aliases"]
                )
            else:
                metadata["albumsort"] = self._swapprefix(metadata, "album")


class LangFunctions:
    """Provides scripting functions swapprefix_lang, delprefix_lang, and title_lang.

    Combinations of languages and their prefixes/minor words are stored after
    they're used at least once.
    """

    prefixes_cache = {}
    minor_words_cache = {}

    def __init__(self):
        all_articles = set(_all_articles | set(article.capitalize() for article in _all_articles))
        all_minor_words = set(_all_articles | _all_other_minor_words)
        self.prefixes_cache[""] = all_articles
        self.minor_words_cache[""] = all_minor_words

    def _format_languages(self, languages):
        """Filters out the languages that are not available.

        It returns a tuple to be used as key for the two cache dictionaries.
        """
        languages = set(lang[:3].lower() for lang in languages)
        return tuple(languages.intersection(_articles_langs))

    def _create_prefixes_list(self, languages = None, is_title = False):
        """Creates a list of all the prefixes or minor words for all the given languages.

        Args:
            languages (set): All the languages to be considered.
            is_title (bool): If true, only articles are included, with a lower case
                             and a capitalized copy for each. This is because the
                             swapprefix and delprefix functions are case sensitive.
                             If false, also prepositions and conjunctions are included,
                             all in lowercase.

        Returns:
            (set): The set of prefixes or minor words.

        The available languages are saved with ISO 639-3 codes.
        """
        prefixes = []
        for language in languages:
            prefixes.extend(_articles[language])
            if is_title:
                prefixes.extend(_other_minor_words[language])
            else:
                prefixes.extend([article.capitalize() for article in _articles[language]])
        return set(prefixes)

    def _title_case(self, text, lower_case_words):
        """Returns the text in titlecase.

        If a word has an apostrophe and the segment to its left is an article,
        it capitalizes only the word on the right. Otherwise it capitalizes
        only the word on the left.
        For example, "let's groove" becomes "Let's Groove", but "voglio l'anima"
        becomes "Voglio l'Anima".
        """
        new_text = []
        words = re.split(r"([\w']+)", text.strip().lower().replace("’", "'"))
        words = [word for word in words if word]
        for word in words:
            if "'" in word:  # Apply the rule described above
                split = word.split("'")
                if split[0] + "'" in lower_case_words:
                    split[1] = split[1].capitalize()
                else:
                    split[0] = split[0].capitalize()
                word = "'".join(split)
            elif word not in lower_case_words:
                word = word.capitalize()
            new_text.append(word)
        if new_text:
            new_text[0] = new_text[0].capitalize()
            return "".join(new_text)
        else:
            return ""

    def find_languages(self, metadata):
        """Finds the languages from the metadata.

        Returns a set with only an empty string if no language information is found.
        Returns None if the languages found are not available.
        """
        languages = (metadata["language"], metadata["_releaselanguage"])
        languages = set(language for language in languages if language)
        if not languages:
            return {""}  # No language information is found.
        languages = languages.intersection(_articles_langs)
        if not languages:
            return None  # None of the languages found are available.
        return languages

    def find_prefixes(self, languages):
        """Returns the list of prefixes for the given languages.

        If the same combination was previously used, it finds the values stored
        to avoid recreating the same list twice.
        """
        if not languages or "" in languages:
            return self.prefixes_cache[""]
        combination = self._format_languages(languages)
        if combination in self.prefixes_cache:
            return self.prefixes_cache[combination]
        prefixes = self._create_prefixes_list(combination)
        self.prefixes_cache[combination] = prefixes
        return prefixes

    def find_minor_words(self, languages):
        """Returns the list of minor words for the given languages.

        If the same combination was previously used, it finds the values stored
        to avoid recreating the same list twice.
        """
        if not languages or "" in languages:
            return self.minor_words_cache[""]
        combination = self._format_languages(languages)
        if combination in self.minor_words_cache:
            return self.minor_words_cache[combination]
        minor_words = self._create_prefixes_list(combination, True)
        self.minor_words_cache[combination] = minor_words
        return minor_words

    swapprefix_lang_documentation = N_(
        """`$swapprefix_lang(text,language1,language2,...)`

Moves the prefix to the end of the text. It uses a list prefixes
taken from the specified languages.
Multiple languages can be added as seperate parameters.
If none are provided, it uses all the available ones.
        """)

    def swapprefix_lang(self, parser, text, *languages):
        """
        >>> lf = LangFunctions()
        >>> lf.swapprefix_lang(None, "The prefix is swapped")
        'prefix is swapped, The'
        >>> lf.swapprefix_lang(None, "the is swapped also in lower case")
        'is swapped also in lower case, the'
        >>> lf.swapprefix_lang(None, "the prefix is not in Spanish", "spa")
        'the prefix is not in Spanish'
        >>> lf.swapprefix_lang(None, "Il prefisso è scambiato", "ita")
        'prefisso è scambiato, Il'
        >>> lf.swapprefix_lang(None, "the-prefix-is-not-swapped")
        'the-prefix-is-not-swapped'
        >>> lf.swapprefix_lang(None, "the")
        'the'
        """
        if parser and not languages:
            languages = self.find_languages(parser.context)
        prefixes = self.find_prefixes(languages)
        return func_swapprefix(parser, text, *prefixes)

    delprefix_lang_documentation = N_(
        """`$delprefix_lang(text,language1,language2,...)`

Deletes the prefix from the text. It uses a list prefixes
taken from the specified languages.
Multiple languages can be added as seperate parameters.
If none are provided, it uses all the available ones.
        """)

    def delprefix_lang(self, parser, text, *languages):
        """
        >>> lf = LangFunctions()
        >>> lf.delprefix_lang(None, "The prefix is deleted")
        'prefix is deleted'
        >>> lf.delprefix_lang(None, "the is deleted also in lower case")
        'is deleted also in lower case'
        >>> lf.delprefix_lang(None, "the prefix is not in Spanish", "spa")
        'the prefix is not in Spanish'
        >>> lf.delprefix_lang(None, "Il prefisso è eliminato", "ita")
        'prefisso è eliminato'
        >>> lf.delprefix_lang(None, "the-prefix-is-not-deleted")
        'the-prefix-is-not-deleted'
        >>> lf.delprefix_lang(None, "the")
        'the'
        """
        if parser and not languages:
            languages = self.find_languages(parser.context)
        prefixes = self.find_prefixes(languages)
        return func_delprefix(parser, text, *prefixes)

    title_lang_documentation = N_(
        """`$title_lang(text,language1,language2,...)`

Makes the text title case based on the minor words of the specified languages.
Multiple languages can be added as separate parameters.
If none are provided, it uses all the available ones.
        """)

    def title_lang(self, parser, text, *languages):
        """
        >>> lf = LangFunctions()
        >>> lf.title_lang(None, "title case")
        'Title Case'
        >>> lf.title_lang(None, "prepositions and/or conjunctions")
        'Prepositions and/or Conjunctions'
        >>> lf.title_lang(None, "(random-punctuation/and.symbols@")
        '(Random-Punctuation/and.Symbols@'
        >>> lf.title_lang(None, "only in english: and, or, but...", "eng")
        'Only in English: and, or, but...'
        >>> lf.title_lang(None, "only in spanish: and, or, but...", "spa")
        'Only In Spanish: And, Or, But...'
        >>> lf.title_lang(None, "Also in other languages, con altre lingue", "eng", "ita")
        'Also in Other Languages, con Altre Lingue'
        >>> lf.title_lang(None, "with no valid language", "aaa")
        'With No Valid Language'
        >>> lf.title_lang(None, "let's groove")
        "Let's Groove"
        >>> lf.title_lang(None, "voglio l'anima")
        "Voglio l'Anima"
        >>> lf.title_lang(None, "casiopea, カシオペア")
        'Casiopea, カシオペア'
        """
        if text.upper() == text and config.setting[KEEP_ALLCAPS]:
            return text
        if parser and not languages:
            languages = self.find_languages(parser.context)
        minor_words = self.find_minor_words(languages)
        return self._title_case(text, minor_words)


class EnhancedTitlesOptions(OptionsPage):
    """Options page found under the "plugins" page.
    """

    NAME = "enhanced_titles"
    TITLE = "Enhanced Titles"
    PARENT = "plugins"

    options = [
        config.BoolOption("setting", KEEP_ALLCAPS, False),
        config.BoolOption("setting", ENABLE_TAGGING, True),
        config.BoolOption("setting", CHECK_ALBUM, True),
        config.BoolOption("setting", CHECK_TRACK, False),
    ]

    def __init__(self, parent = None):
        super(EnhancedTitlesOptions, self).__init__(parent)
        self.ui = Ui_EnhancedTitlesOptions()
        self.ui.setupUi(self)

    def load(self):
        self.ui.check_allcaps.setChecked(config.setting[KEEP_ALLCAPS])
        self.ui.check_tagging.setChecked(config.setting[ENABLE_TAGGING])
        self.ui.check_album_aliases.setChecked(config.setting[CHECK_ALBUM])
        self.ui.check_track_aliases.setChecked(config.setting[CHECK_TRACK])

    def save(self):
        config.setting[KEEP_ALLCAPS] = self.ui.check_allcaps.isChecked()
        config.setting[ENABLE_TAGGING] = self.ui.check_tagging.isChecked()
        config.setting[CHECK_ALBUM] = self.ui.check_album_aliases.isChecked()
        config.setting[CHECK_TRACK] = self.ui.check_track_aliases.isChecked()


sort_tagger = SortTagger()
lang_functions = LangFunctions()
register_track_metadata_processor(sort_tagger.set_track_titlesort, priority = PluginPriority.LOW)
register_album_metadata_processor(sort_tagger.set_album_titlesort, priority = PluginPriority.LOW)
register_script_function(lang_functions.swapprefix_lang, check_argcount = False, documentation = lang_functions.swapprefix_lang_documentation)
register_script_function(lang_functions.delprefix_lang, check_argcount = False, documentation = lang_functions.delprefix_lang_documentation)
register_script_function(lang_functions.title_lang, check_argcount = False, documentation = lang_functions.title_lang_documentation)
register_options_page(EnhancedTitlesOptions)



# -*- coding: utf-8 -*-
PLUGIN_NAME = "ReplayGain 2.0"
PLUGIN_AUTHOR = "complexlogic"
PLUGIN_DESCRIPTION = '''
Calculates ReplayGain information for tracks and albums according to the
[ReplayGain 2.0 specification](https://wiki.hydrogenaud.io/index.php?title=ReplayGain_2.0_specification).
This plugin depends on the ReplayGain utility [rsgain](https://github.com/complexlogic/rsgain). Users
are required to install rsgain and set its path in the plugin settings before use.

#### Usage
Select one or more tracks or albums, then right click and select Plugin->Calculate ReplayGain. The plugin
will calculate ReplayGain information for the selected items and display the results in the metadata
window. Click the save button to write the tags to file.

The following file formats are supported:

- MP3 (.mp3)
- FLAC (.flac)
- Ogg (.ogg, .oga, spx)
- Opus (.opus)
- MPEG-4 Audio (.m4a, .mp4)
- Wavpack (.wv)
- Monkey's Audio (.ape)
- WMA (.wma)
- MP2 (.mp2)
- WAV (.wav)
- AIFF (.aiff)
- TAK (.tak)
- MusePack *(Stream Version 8 only)* (.mpc)

This plugin is based on the original ReplayGain plugin by Philipp Wolfer and Sophist.
'''
PLUGIN_VERSION = "1.7.1"
PLUGIN_API_VERSIONS = ["2.0"]
PLUGIN_LICENSE = "GPL-2.0"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.html"

from enum import IntEnum
from functools import partial
import subprocess  # nosec: B404
import shutil
import os

from PyQt5.QtWidgets import QFileDialog

from picard import log
from picard.formats import (
    AiffFile,
    ASFFile,
    FLACFile,
    MonkeysAudioFile,
    MP3File,
    MP4File,
    MusepackFile,
    OggFLACFile,
    OggOpusFile,
    OggSpeexFile,
    OggTheoraFile,
    OggVorbisFile,
    TAKFile,
    WAVFile,
    WavPackFile,
)
from picard.album import Album
from picard.track import Track, NonAlbumTrack
from picard.util import thread
from picard.ui.options import register_options_page, OptionsPage
from picard.config import TextOption, BoolOption, IntOption, config
from picard.ui.itemviews import (BaseAction, register_track_action,
                                 register_album_action)
from picard.plugins.replaygain2.ui_options_replaygain2 import Ui_ReplayGain2OptionsPage


CLIP_MODE_DISABLED = 0
CLIP_MODE_POSITIVE = 1
CLIP_MODE_ALWAYS = 2

CLIP_MODE_MAP = (
    "n",
    "p",
    "a"
)

SUPPORTED_FORMATS = (
    AiffFile,
    ASFFile,
    FLACFile,
    MonkeysAudioFile,
    MP3File,
    MP4File,
    MusepackFile,
    OggFLACFile,
    OggOpusFile,
    OggSpeexFile,
    OggTheoraFile,
    OggVorbisFile,
    TAKFile,
    WAVFile,
    WavPackFile,
)

TABLE_HEADER = (
    "filename",
    "loudness",
    "gain",
    "peak",
    "peak_db",
    "peak_type",
    "clipping_adjustment"
)

TAGS = (
    "replaygain_album_gain",
    "replaygain_album_peak",
    "replaygain_album_range",
    "replaygain_reference_loudness",
    "replaygain_track_gain",
    "replaygain_track_peak",
    "replaygain_track_range",
    "r128_album_gain",
    "r128_track_gain"
)


class OpusMode(IntEnum):
    STANDARD = 0
    R128 = 1
    BOTH = 2


# Make sure the rsgain executable exists
def rsgain_found(rsgain_command, window):
    if not os.path.exists(rsgain_command) and shutil.which(rsgain_command) is None:
        window.set_statusbar_message("Failed to locate rsgain. Enter the path in the plugin settings.")
        return False
    return True

# Convert Picard settings dict to rsgain command line options
def build_options(config):
    options = ["custom", "-O", "-s", "s"]
    if (config.setting["album_tags"]):
        options.append("-a")
    if (config.setting["true_peak"]):
        options.append("-t")
    options += ["-l", str(config.setting["target_loudness"])]
    options += ["-c", CLIP_MODE_MAP[config.setting["clip_mode"]]]
    options += ["-m", str(config.setting["max_peak"])]
    return options

# Convert table row to result dict
def parse_result(line):
    result = dict()
    columns = line.split("\t")

    if len(columns) != len(TABLE_HEADER):
        return None
    for i, column in enumerate(columns):
        result[TABLE_HEADER[i]] = column
    return result

# Format the gain as a Q7.8 fixed point number per RFC 7845
# see: https://datatracker.ietf.org/doc/html/rfc7845
def format_r128(result, config):
    gain = float(result["gain"])
    if config.setting["opus_m23"]:
        gain += float(-23 - config.setting["target_loudness"])
    return str(int(round(gain * 256.0)))

def update_metadata(metadata, track_result, album_result, is_nat, opus_mode):
    for tag in TAGS:
        metadata.delete(tag)

    # Opus R128 tags
    if opus_mode in (OpusMode.R128, OpusMode.BOTH):
        metadata.set("r128_track_gain", format_r128(track_result, config))
        if album_result is not None:
            metadata.set("r128_album_gain", format_r128(album_result, config))

    # Standard ReplayGain tags
    if opus_mode in (OpusMode.STANDARD, OpusMode.BOTH):
        metadata.set("replaygain_track_gain", track_result["gain"] + " dB")
        metadata.set("replaygain_track_peak", track_result["peak"])
        if config.setting["album_tags"]:
            if is_nat:
                metadata.set("replaygain_album_gain", track_result["gain"] + " dB")
                metadata.set("replaygain_album_peak", track_result["peak"])
            elif album_result is not None:
                metadata.set("replaygain_album_gain", album_result["gain"] + " dB")
                metadata.set("replaygain_album_peak", album_result["peak"])
        if config.setting["reference_loudness"]:
            metadata.set("replaygain_reference_loudness", f"{float(config.setting['target_loudness']):.2f} LUFS")

def calculate_replaygain(tracks, options):

    # Make sure files are of supported type, build file list
    files = list()
    valid_tracks = list()
    for track in tracks:
        if not track.files:
            continue
        file = track.files[0]
        if not isinstanceany(file, SUPPORTED_FORMATS):
            raise Exception(f"ReplayGain 2.0: File '{file.filename}' is of unsupported format")
        files.append(file.filename)
        valid_tracks.append(track)

    call = [config.setting["rsgain_command"]] + options + files
    for item in call:
        item.encode("utf-8")

    # Prevent an unwanted console spawn in Windows
    si = None
    if (os.name == "nt"):
        si = subprocess.STARTUPINFO()
        si.dwFlags = subprocess.STARTF_USESHOWWINDOW
        si.wShowWindow = subprocess.SW_HIDE

    # Execute the scan with rsgain
    lines = list()
    with subprocess.Popen(  # nosec: B603
        call,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        startupinfo=si,
        encoding="utf-8",
        text=True
    ) as process:
        (output, _unused) = process.communicate()
        rc = process.poll()
        if rc:
            raise Exception(f'ReplayGain 2.0: rsgain returned non-zero code ({rc})')
        log.debug(output)
        lines = output.splitlines()
    album_tags = config.setting["album_tags"]

    # Make sure the number of rows in the output is what we expected
    if (len(lines) !=
        1                            # Table header
        + len(valid_tracks)          # 1 row per track
        + 1 if album_tags else 0):   # Album result
        raise Exception(f"ReplayGain 2.0: Unexpected output from rsgain: {lines}")
    lines.pop(0) # Don't care about the table header

    # Parse album result
    album_result = None
    if album_tags:
        album_result = parse_result(lines[-1])
        lines.pop(-1)

    # Parse track results
    results = list()
    for line in lines:
        result = parse_result(line)
        if result is None:
            raise Exception("ReplayGain 2.0: Failed to parse result")
        results.append(result)

    # Update track metadata with results
    if isinstance(file, OggOpusFile):
        opus_mode = config.setting["opus_mode"]
    else:
        opus_mode = OpusMode.STANDARD

    for i, track in enumerate(valid_tracks):
        for file in track.files:
            update_metadata(
                file.metadata,
                results[i],
                album_result,
                isinstance(track, NonAlbumTrack),
                opus_mode
            )


def isinstanceany(obj, types):
    return any(isinstance(obj, t) for t in types)


class ScanTracks(BaseAction):
    NAME = "Calculate Replay&Gain..."

    def callback(self, objs):
        if not rsgain_found(self.config.setting["rsgain_command"], self.tagger.window):
            return
        tracks = list(filter(lambda o: isinstance(o, Track), objs))
        self.options = build_options(self.config)
        num_tracks = len(tracks)
        if num_tracks == 1:
            self.tagger.window.set_statusbar_message(
                'Calculating ReplayGain for %s...', {tracks[0].files[0].filename}
            )
        else:
            self.tagger.window.set_statusbar_message(
                'Calculating ReplayGain for %i tracks...', num_tracks
            )
        thread.run_task(
            partial(calculate_replaygain, tracks, self.options),
            partial(self._replaygain_callback, tracks)
        )

    def _replaygain_callback(self, tracks, result=None, error=None):
        if error is None:
            for track in tracks:
                for file in track.files:
                    file.update()
                track.update()
            self.tagger.window.set_statusbar_message('ReplayGain successfully calculated.')
        else:
            self.tagger.window.set_statusbar_message('Could not calculate ReplayGain.')


class ScanAlbums(BaseAction):
    NAME = "Calculate Replay&Gain..."

    def callback(self, objs):
        if not rsgain_found(self.config.setting["rsgain_command"], self.tagger.window):
            return
        self.options = build_options(self.config)
        albums = list(filter(lambda o: isinstance(o, Album), objs))

        self.num_albums = len(albums)
        self.current = 0
        if (self.num_albums == 1):
            self.tagger.window.set_statusbar_message(
                'Calculating ReplayGain for "%s"...', albums[0].metadata['album']
            )
        else:
            self.tagger.window.set_statusbar_message(
                'Calculating ReplayGain for %i albums...', self.num_albums
            )
        for album in albums:
            thread.run_task(
                partial(calculate_replaygain, album.tracks, self.options),
                partial(self._albumgain_callback, album)
            )

    def _format_progress(self):
        if self.num_albums == 1:
            return ""
        else:
            self.current += 1
            return f" ({self.current}/{self.num_albums})"

    def _albumgain_callback(self, album, result=None, error=None):
        progress = self._format_progress()
        if error is None:
            for track in album.tracks:
                for file in track.files:
                    file.update()
                track.update()
            album.update()
            self.tagger.window.set_statusbar_message(
                'Successfully calculated ReplayGain for "%(album)s"%(progress)s.',
                {
                    'album': album.metadata['album'],
                    'progress': progress,
                }
            )
        else:
            self.tagger.window.set_statusbar_message(
                'Failed to calculate ReplayGain for "%(album)s"%(progress)s.',
                {
                    'album': album.metadata['album'],
                    'progress': progress,
                }
            )

class ReplayGain2OptionsPage(OptionsPage):

    NAME = "replaygain2"
    TITLE = "ReplayGain 2.0"
    PARENT = "plugins"

    options = [
        TextOption("setting", "rsgain_command", "rsgain"),
        BoolOption("setting", "album_tags", True),
        BoolOption("setting", "true_peak", False),
        BoolOption("setting", "reference_loudness", False),
        IntOption("setting", "target_loudness", -18),
        IntOption("setting", "clip_mode", CLIP_MODE_POSITIVE),
        IntOption("setting", "max_peak", 0),
        IntOption("setting", "opus_mode", OpusMode.STANDARD),
        BoolOption("setting", "opus_m23", False)
    ]

    def __init__(self, parent=None):
        super(ReplayGain2OptionsPage, self).__init__(parent)
        self.ui = Ui_ReplayGain2OptionsPage()
        self.ui.setupUi(self)
        self.ui.clip_mode.addItems([
            "Disabled",
            "Enabled for positive gain values only",
            "Always enabled"
        ])
        self.ui.opus_mode.addItems([
            "Write standard ReplayGain tags",
            "Write R128_*_GAIN tags",
            "Write both standard and R128 tags"
        ])
        self.ui.rsgain_command_browse.clicked.connect(self.rsgain_command_browse)

    def load(self):
        self.ui.rsgain_command.setText(self.config.setting["rsgain_command"])
        self.ui.album_tags.setChecked(self.config.setting["album_tags"])
        self.ui.true_peak.setChecked(self.config.setting["true_peak"])
        self.ui.reference_loudness.setChecked(self.config.setting["reference_loudness"])
        self.ui.target_loudness.setValue(self.config.setting["target_loudness"])
        self.ui.clip_mode.setCurrentIndex(self.config.setting["clip_mode"])
        self.ui.max_peak.setValue(self.config.setting["max_peak"])
        self.ui.opus_mode.setCurrentIndex(self.config.setting["opus_mode"])
        self.ui.opus_m23.setChecked(self.config.setting["opus_m23"])

    def save(self):
        self.config.setting["rsgain_command"] = self.ui.rsgain_command.text()
        self.config.setting["album_tags"] = self.ui.album_tags.isChecked()
        self.config.setting["true_peak"] = self.ui.true_peak.isChecked()
        self.config.setting["reference_loudness"] = self.ui.reference_loudness.isChecked()
        self.config.setting["target_loudness"] = self.ui.target_loudness.value()
        self.config.setting["clip_mode"] = self.ui.clip_mode.currentIndex()
        self.config.setting["max_peak"] = self.ui.max_peak.value()
        self.config.setting["opus_mode"] = self.ui.opus_mode.currentIndex()
        self.config.setting["opus_m23"] = self.ui.opus_m23.isChecked()

    def rsgain_command_browse(self):
        path, _filter = QFileDialog.getOpenFileName(self, "", self.ui.rsgain_command.text())
        if path:
            path = os.path.normpath(path)
            self.ui.rsgain_command.setText(path)


register_track_action(ScanTracks())
register_album_action(ScanAlbums())
register_options_page(ReplayGain2OptionsPage)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2017, 2019, 2021 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

PLUGIN_NAME = "Load as non-album track"
PLUGIN_AUTHOR = "Philipp Wolfer"
PLUGIN_DESCRIPTION = ("Allows loading selected tracks as non-album tracks. "
                      "Useful for tagging single tracks where you do not care "
                      "about the album.")
PLUGIN_VERSION = "0.4"
PLUGIN_API_VERSIONS = ["1.4.0", "2.0", "2.1", "2.2"]
PLUGIN_LICENSE = "GPL-2.0-or-later"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.html"

from picard import log
from picard.track import Track
from picard.ui.itemviews import (
    BaseAction,
    register_track_action,
)


class LoadAsNat(BaseAction):
    NAME = "Load as non-album track..."

    def callback(self, objs):
        for track in (t for t in objs if isinstance(t, Track)):
            nat = self.tagger.load_nat(
                track.metadata['musicbrainz_recordingid'])
            for file in list(track.linked_files):
                file.move(nat)
                metadata = file.metadata
                metadata.delete('albumartist')
                metadata.delete('albumartistsort')
                metadata.delete('albumsort')
                metadata.delete('asin')
                metadata.delete('barcode')
                metadata.delete('catalognumber')
                metadata.delete('discnumber')
                metadata.delete('discsubtitle')
                metadata.delete('media')
                metadata.delete('musicbrainz_albumartistid')
                metadata.delete('musicbrainz_albumid')
                metadata.delete('musicbrainz_discid')
                metadata.delete('musicbrainz_releasegroupid')
                metadata.delete('releasecountry')
                metadata.delete('releasestatus')
                metadata.delete('releasetype')
                metadata.delete('totaldiscs')
                metadata.delete('totaltracks')
                metadata.delete('tracknumber')
                log.debug("[LoadAsNat] deleted tags: %r", metadata.deleted_tags)


register_track_action(LoadAsNat())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# This is the Smart Title Case plugin for MusicBrainz Picard.
# Copyright (C) 2017 Sophist.
#
# Updated for use with Picard v2.0 by Bob Swift (rdswift).
#
# It is based on the Title Case plugin by Javier Kohen
# Copyright 2007 Javier Kohen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

PLUGIN_NAME = "Smart Title Case"
PLUGIN_AUTHOR = "Sophist based on an earlier plugin by Javier Kohen"
PLUGIN_DESCRIPTION = """
Capitalize First Character In Every Word Of Album/Track Title/Artist.<br />
Leaves words containing embedded uppercase as-is i.e. USA or DoA.<br />
For Artist/AlbumArtist, title cases only artists not join phrases<br />
e.g. The Beatles feat. The Who.
"""
PLUGIN_VERSION = "0.4.2"
PLUGIN_API_VERSIONS = ["2.0"]
PLUGIN_LICENSE = "GPL-2.0-or-later"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-3.0.html"

import re, unicodedata

from picard import log
from picard.metadata import (
    register_track_metadata_processor,
    register_album_metadata_processor,
)


title_tags = ['title', 'album']
artist_tags = [
    ('artist', 'artists'),
    ('artistsort', '~artists_sort'),
    ('albumartist', '~albumartists'),
    ('albumartistsort', '~albumartists_sort'),
    ]
title_re = re.compile(r'\w[^-,/\s\u2010\u2011]*')

def match_word(match):
    word = match.group(0)
    if word == word.lower():
        word = word[0].upper() + word[1:]
    return word

def string_title_match(match_word, string):
    return title_re.sub(match_word, string)

def string_cleanup(string):
    if not string:
        return ""
    return unicodedata.normalize("NFKC", string)

def string_title_case(string):
    """Title-case a string using a less destructive method than str.title.
    >>> string_title_case('make title case')
    'Make Title Case'
    >>> string_title_case('Already Title Case')
    'Already Title Case'
    >>> string_title_case('mIxEd cAsE')
    'mIxEd cAsE'
    >>> string_title_case('a')
    'A'
    >>> string_title_case("apostrophe's apostrophe's")
    "Apostrophe's Apostrophe's"
    >>> string_title_case('(bracketed text)')
    '(Bracketed Text)'
    >>> string_title_case("'single quotes'")
    "'Single Quotes'"
    >>> string_title_case('"double quotes"')
    '"Double Quotes"'
    >>> string_title_case('a,b')
    'A,B'
    >>> string_title_case('a-b')
    'A-B'
    >>> string_title_case('a/b')
    'A/B'
    >>> string_title_case('flügel')
    'Flügel'
    >>> string_title_case('HARVEST STORY by 杉山清貴')
    'HARVEST STORY By 杉山清貴'
    """
    return string_title_match(match_word, string_cleanup(string))


def artist_title_case(text, artists, artists_upper):
    """
    Use the array of artists and the joined string
    to identify artists to make title case
    and the join strings to leave as-is.

    >>> artist_title_case('the beatles feat. the who', ['the beatles', 'the who'], ['The Beatles', 'The Who'])
    'The Beatles feat. The Who'
    >>> artist_title_case('kesha feat. 3OH!3', ['kesha', '3OH!3'], ['Kesha', '3OH!3'])
    'Kesha feat. 3OH!3'
    """
    find = "^(" + r")(\s+\S+?\s+)(".join((map(re.escape, map(string_cleanup,artists)))) + ")(.*$)"
    replace = "".join([r"%s\g<%d>" % (a, x*2 + 2) for x, a in enumerate(artists_upper)])
    result = re.sub(find, replace, string_cleanup(text))
    return result


def title_case(tagger, metadata, *args):
    for name in title_tags:
        if name in metadata:
            values = metadata.getall(name)
            new_values = [string_title_case(v) for v in values]
            if values != new_values:
                log.debug("SmartTitleCase: %s: %r replaced with %r", name, values, new_values)
                metadata[name] = new_values
    for artist_string, artists_list in artist_tags:
        if artist_string in metadata and artists_list in metadata:
            artist = metadata.getall(artist_string)
            artists = metadata.getall(artists_list)
            new_artists = list(map(string_title_case, artists))
            new_artist = [artist_title_case(x, artists, new_artists) for x in artist]
            if artists != new_artists and artist != new_artist:
                log.debug("SmartTitleCase: %s: %s replaced with %s", artist_string, artist, new_artist)
                log.debug("SmartTitleCase: %s: %r replaced with %r", artists_list, artists, new_artists)
                metadata[artist_string] = new_artist
                metadata[artists_list] = new_artists
            elif artists != new_artists or artist != new_artist:
                if artists != new_artists:
                    log.warning("SmartTitleCase: %s changed, %s wasn't", artists_list, artist_string)
                    log.warning("SmartTitleCase: %s: %r changed to %r", artists_list, artists, new_artists)
                    log.warning("SmartTitleCase: %s: %r unchanged", artist_string, artist)
                else:
                    log.warning("SmartTitleCase: %s changed, %s wasn't", artist_string, artists_list)
                    log.warning("SmartTitleCase: %s: %r changed to %r", artist_string, artist, new_artist)
                    log.warning("SmartTitleCase: %s: %r unchanged", artists_list, artists)

register_track_metadata_processor(title_case)
register_album_metadata_processor(title_case)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2015-2020 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

PLUGIN_NAME = 'TheAudioDB cover art'
PLUGIN_AUTHOR = 'Philipp Wolfer'
PLUGIN_DESCRIPTION = 'Use cover art from TheAudioDB.'
PLUGIN_VERSION = "1.3.1"
PLUGIN_API_VERSIONS = ["2.0", "2.1", "2.2", "2.3", "2.4", "2.5", "2.6"]
PLUGIN_LICENSE = "GPL-2.0-or-later"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.html"

from base64 import b64decode
from PyQt5.QtCore import QUrl
from PyQt5.QtNetwork import QNetworkReply
from picard import config, log
from picard.coverart.providers import (
    CoverArtProvider,
    ProviderOptions,
    register_cover_art_provider,
)
from picard.coverart.image import CoverArtImage
from picard.config import (
    BoolOption,
    TextOption,
)
from picard.webservice import ratecontrol
from .ui_options_theaudiodb import Ui_TheAudioDbOptionsPage

THEAUDIODB_HOST = "www.theaudiodb.com"
THEAUDIODB_PORT = 443
THEAUDIODB_APIKEY = 'MWQ2NTY1NjQ2OTRmMTM0ZDY1NjU2NA=='

OPTION_CDART_ALWAYS = "always"
OPTION_CDART_NEVER = "never"
OPTION_CDART_NOALBUMART = "noalbumart"


# No rate limit for TheAudioDB.
ratecontrol.set_minimum_delay((THEAUDIODB_HOST, THEAUDIODB_PORT), 0)


class TheAudioDbOptionsPage(ProviderOptions):

    _options_ui = Ui_TheAudioDbOptionsPage

    options = [
        TextOption("setting", "theaudiodb_use_cdart", OPTION_CDART_NOALBUMART),
        BoolOption("setting", "theaudiodb_use_high_quality", False),
    ]

    def load(self):
        if config.setting["theaudiodb_use_cdart"] == OPTION_CDART_ALWAYS:
            self.ui.theaudiodb_cdart_use_always.setChecked(True)
        elif config.setting["theaudiodb_use_cdart"] == OPTION_CDART_NEVER:
            self.ui.theaudiodb_cdart_use_never.setChecked(True)
        elif config.setting["theaudiodb_use_cdart"] == OPTION_CDART_NOALBUMART:
            self.ui.theaudiodb_cdart_use_if_no_albumcover.setChecked(True)
        self.ui.theaudiodb_use_high_quality.setChecked(
            config.setting['theaudiodb_use_high_quality'])

    def save(self):
        if self.ui.theaudiodb_cdart_use_always.isChecked():
            config.setting["theaudiodb_use_cdart"] = OPTION_CDART_ALWAYS
        elif self.ui.theaudiodb_cdart_use_never.isChecked():
            config.setting["theaudiodb_use_cdart"] = OPTION_CDART_NEVER
        elif self.ui.theaudiodb_cdart_use_if_no_albumcover.isChecked():
            config.setting["theaudiodb_use_cdart"] = OPTION_CDART_NOALBUMART
        config.setting['theaudiodb_use_high_quality'] = self.ui.theaudiodb_use_high_quality.isChecked()


class TheAudioDbCoverArtImage(CoverArtImage):

    """Image from The Audio DB"""

    support_types = True
    sourceprefix = "AUDIODB"

    def parse_url(self, url):
        super().parse_url(url)
        # Workaround for Picard always returning port 80 regardless of the
        # scheme. No longer necessary for Picard >= 2.1.3
        self.port = self.url.port(443 if self.url.scheme() == 'https' else 80)


class CoverArtProviderTheAudioDb(CoverArtProvider):

    """Use TheAudioDB to get cover art"""

    NAME = "TheAudioDB"
    TITLE = "TheAudioDB"
    OPTIONS = TheAudioDbOptionsPage

    def __init__(self, coverart):
        super().__init__(coverart)
        self.__api_key = b64decode(THEAUDIODB_APIKEY).decode()

    def enabled(self):
        return super().enabled() and not self.coverart.front_image_found

    def queue_images(self):
        release_group_id = self.metadata["musicbrainz_releasegroupid"]
        path = "/api/v1/json/%s/album-mb.php" % self.__api_key
        queryargs = {
            "i": bytes(QUrl.toPercentEncoding(release_group_id)).decode()
        }
        log.debug("TheAudioDB: Queued download: %s?i=%s", path, queryargs["i"])
        self.album.tagger.webservice.get(
            THEAUDIODB_HOST,
            THEAUDIODB_PORT,
            path,
            self._json_downloaded,
            priority=True,
            important=False,
            parse_response_type='json',
            queryargs=queryargs)
        self.album._requests += 1
        return CoverArtProvider.WAIT

    def _json_downloaded(self, data, reply, error):
        self.album._requests -= 1

        if error:
            if error != QNetworkReply.ContentNotFoundError:
                error_level = log.error
            else:
                error_level = log.debug
            error_level("TheAudioDB: Problem requesting metadata: %s", error)
        else:
            try:
                releases = data.get("album")
                if not releases:
                    log.debug("TheAudioDB: No cover art found for %s",
                              reply.url().url())
                    return
                release = releases[0]
                albumart_url = None
                if config.setting['theaudiodb_use_high_quality']:
                    albumart_url = release.get("strAlbumThumbHQ")
                if not albumart_url:
                    albumart_url = release.get("strAlbumThumb")
                cdart_url = release.get("strAlbumCDart")
                use_cdart = config.setting["theaudiodb_use_cdart"]

                if albumart_url:
                    self._select_and_add_cover_art(albumart_url, ["front"])

                if cdart_url and (use_cdart == OPTION_CDART_ALWAYS
                                  or (use_cdart == OPTION_CDART_NOALBUMART
                                      and not albumart_url)):
                    types = ["medium"]
                    if not albumart_url:
                        types.append("front")
                    self._select_and_add_cover_art(cdart_url, types)
            except (TypeError):
                log.error("TheAudioDB: Problem processing downloaded metadata", exc_info=True)
            finally:
                self.next_in_queue()

    def _select_and_add_cover_art(self, url, types):
        log.debug("TheAudioDB: Found artwork %s" % url)
        self.queue_put(TheAudioDbCoverArtImage(url, types=types))


register_cover_art_provider(CoverArtProviderTheAudioDb)



PLUGIN_NAME = 'Musixmatch Lyrics'
PLUGIN_AUTHOR = 'm-yn, Sambhav Kothari, Philipp Wolfer'
PLUGIN_DESCRIPTION = 'Fetch first 30% of lyrics from Musixmatch'
PLUGIN_VERSION = '1.1.1'
PLUGIN_API_VERSIONS = ["2.0"]
PLUGIN_LICENSE = "GPL-2.0"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.html"


from functools import partial
from picard import config, log
from picard.metadata import register_track_metadata_processor
from picard.ui.options import register_options_page, OptionsPage
from .ui_options_musixmatch import Ui_MusixmatchOptionsPage


MUSIXMATCH_HOST = 'api.musixmatch.com'
MUSIXMATCH_PORT = 80


def handle_result(album, metadata, data, reply, error):
    try:
        if error:
            log.error(error)
            return
        message = data.get('message', {})
        header = message.get('header')
        if header.get('status_code') != 200:
            log.warning('MusixMatch: Server returned no result: %s', data)
            return
        result = message.get('body', {}).get('lyrics')
        if result:
            lyrics = result.get('lyrics_body')
            if lyrics:
                metadata['lyrics:description'] = lyrics
    except AttributeError:
        log.error('MusixMatch: Error handling server response %s',
                  data, exc_info=True)
    finally:
        album._requests -= 1
        album._finalize_loading(error)


def process_track(album, metadata, track, release):
    apikey = config.setting['musixmatch_api_key']
    if not apikey:
        log.warning('MusixMatch: No API key configured')
        return
    if metadata['language'] == 'zxx':
        log.debug('MusixMatch: Track %s has no lyrics, skipping query',
                  metadata['musicbrainz_recordingid'])
        return
    queryargs = {
        'apikey': apikey,
        'track_mbid': metadata['musicbrainz_recordingid']
    }
    album.tagger.webservice.get(
        MUSIXMATCH_HOST,
        MUSIXMATCH_PORT,
        "/ws/1.1/track.lyrics.get",
        partial(handle_result, album, metadata),
        parse_response_type='json',
        queryargs=queryargs
    )
    album._requests += 1


class MusixmatchOptionsPage(OptionsPage):
    NAME = 'musixmatch'
    TITLE = 'Musixmatch API Key'
    PARENT = "plugins"
    options = [
        config.TextOption("setting", "musixmatch_api_key", "")
    ]

    def __init__(self, parent=None):
        super(MusixmatchOptionsPage, self).__init__(parent)
        self.ui = Ui_MusixmatchOptionsPage()
        self.ui.setupUi(self)

    def load(self):
        self.ui.api_key.setText(config.setting["musixmatch_api_key"])

    def save(self):
        self.config.setting["musixmatch_api_key"] = self.ui.api_key.text()


register_track_metadata_processor(process_track)
register_options_page(MusixmatchOptionsPage)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'plugins/search_engine_lookup/ui_options_search_engine_lookup.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_SearchEngineLookupOptionsPage(object):
    def setupUi(self, SearchEngineLookupOptionsPage):
        SearchEngineLookupOptionsPage.setObjectName("SearchEngineLookupOptionsPage")
        SearchEngineLookupOptionsPage.resize(561, 802)
        SearchEngineLookupOptionsPage.setMinimumSize(QtCore.QSize(360, 0))
        self.verticalLayout = QtWidgets.QVBoxLayout(SearchEngineLookupOptionsPage)
        self.verticalLayout.setObjectName("verticalLayout")
        self.scrollArea = QtWidgets.QScrollArea(SearchEngineLookupOptionsPage)
        self.scrollArea.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 543, 784))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.gb_description = QtWidgets.QGroupBox(self.scrollAreaWidgetContents)
        self.gb_description.setObjectName("gb_description")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.gb_description)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.format_description = QtWidgets.QLabel(self.gb_description)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.format_description.sizePolicy().hasHeightForWidth())
        self.format_description.setSizePolicy(sizePolicy)
        self.format_description.setWordWrap(True)
        self.format_description.setObjectName("format_description")
        self.verticalLayout_3.addWidget(self.format_description)
        self.verticalLayout_2.addWidget(self.gb_description)
        self.gb_search_engine = QtWidgets.QGroupBox(self.scrollAreaWidgetContents)
        self.gb_search_engine.setMinimumSize(QtCore.QSize(0, 0))
        self.gb_search_engine.setObjectName("gb_search_engine")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.gb_search_engine)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.list_providers = QtWidgets.QListWidget(self.gb_search_engine)
        self.list_providers.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.list_providers.setObjectName("list_providers")
        self.verticalLayout_4.addWidget(self.list_providers)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setContentsMargins(-1, -1, -1, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.pb_add = QtWidgets.QPushButton(self.gb_search_engine)
        self.pb_add.setObjectName("pb_add")
        self.horizontalLayout.addWidget(self.pb_add)
        self.pb_edit = QtWidgets.QPushButton(self.gb_search_engine)
        self.pb_edit.setObjectName("pb_edit")
        self.horizontalLayout.addWidget(self.pb_edit)
        self.pb_delete = QtWidgets.QPushButton(self.gb_search_engine)
        self.pb_delete.setObjectName("pb_delete")
        self.horizontalLayout.addWidget(self.pb_delete)
        self.pb_test = QtWidgets.QPushButton(self.gb_search_engine)
        self.pb_test.setObjectName("pb_test")
        self.horizontalLayout.addWidget(self.pb_test)
        self.verticalLayout_4.addLayout(self.horizontalLayout)
        self.verticalLayout_2.addWidget(self.gb_search_engine)
        self.groupBox = QtWidgets.QGroupBox(self.scrollAreaWidgetContents)
        self.groupBox.setMinimumSize(QtCore.QSize(0, 0))
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.groupBox)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.label = QtWidgets.QLabel(self.groupBox)
        self.label.setWordWrap(True)
        self.label.setObjectName("label")
        self.verticalLayout_5.addWidget(self.label)
        self.le_additional_words = QtWidgets.QLineEdit(self.groupBox)
        self.le_additional_words.setObjectName("le_additional_words")
        self.verticalLayout_5.addWidget(self.le_additional_words)
        self.verticalLayout_2.addWidget(self.groupBox)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_2.addItem(spacerItem1)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.verticalLayout.addWidget(self.scrollArea)

        self.retranslateUi(SearchEngineLookupOptionsPage)
        QtCore.QMetaObject.connectSlotsByName(SearchEngineLookupOptionsPage)

    def retranslateUi(self, SearchEngineLookupOptionsPage):
        _translate = QtCore.QCoreApplication.translate
        SearchEngineLookupOptionsPage.setWindowTitle(_translate("SearchEngineLookupOptionsPage", "Form"))
        self.gb_description.setTitle(_translate("SearchEngineLookupOptionsPage", "Search Engine Lookup"))
        self.format_description.setText(_translate("SearchEngineLookupOptionsPage", "<html><head/><body><p>The Search Engine Lookup plugin allows you to initiate a search engine lookup in your browser for a cluster from the menu displayed when right-clicking the cluster.  These settings allow you select which search engine to use for the lookup, as well as any additional search terms.</p></body></html>"))
        self.gb_search_engine.setTitle(_translate("SearchEngineLookupOptionsPage", "Search Engine Providers"))
        self.pb_add.setText(_translate("SearchEngineLookupOptionsPage", "Add"))
        self.pb_edit.setText(_translate("SearchEngineLookupOptionsPage", "Edit"))
        self.pb_delete.setText(_translate("SearchEngineLookupOptionsPage", "Delete"))
        self.pb_test.setText(_translate("SearchEngineLookupOptionsPage", "Test"))
        self.groupBox.setTitle(_translate("SearchEngineLookupOptionsPage", "Additional Search Words"))
        self.label.setText(_translate("SearchEngineLookupOptionsPage", "<html><head/><body><p>By default, the search parameters used are the artist name and album name if they are available in the metadata. You have the option of specifying additional words to be added to the search parameters (e.g.: album).</p><p>Additional words are entered below and should be separated by spaces.</p></body></html>"))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 Bob Swift (rdswift)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

PLUGIN_NAME = 'Search Engine Lookup'
PLUGIN_AUTHOR = 'Bob Swift'
PLUGIN_DESCRIPTION = '''
Adds a right click option on a cluster to look up album information using a search engine in a browser window.

Adds a right click option on an album or track to look up cover art for the selected album or title.
'''
PLUGIN_VERSION = '2.1.0'
PLUGIN_API_VERSIONS = ['2.0', '2.1', '2.2', '2.3']
PLUGIN_LICENSE = "GPL-2.0"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.txt"

import re
from urllib.parse import quote_plus
from uuid import uuid4

from PyQt5 import QtCore, QtWidgets

from picard import config, log
from picard.cluster import Cluster
from picard.plugins.search_engine_lookup.ui_options_search_engine_editor import Ui_SearchEngineEditorDialog
from picard.plugins.search_engine_lookup.ui_options_search_engine_lookup import Ui_SearchEngineLookupOptionsPage
from picard.ui.itemviews import BaseAction, register_cluster_action, register_album_action, register_track_action
from picard.ui.options import OptionsPage, register_options_page
from picard.util.webbrowser2 import open as _open

DEFAULT_PROVIDERS = {
    'ea520f49-36bc-4821-a16a-38bf0340d1f3': {'name': _('Google'), 'url': r'https://www.google.com/search?q=%search%'},
    '7b93d4b5-34d9-49a7-901c-ed0914f07aee': {'name': _('Bing'), 'url': r'https://www.bing.com/search?q=%search%'},
    '37be75d9-5fc5-4858-87fc-b5db0896a163': {'name': _('DuckDuckGo'), 'url': r'https://duckduckgo.com/?q=%search%'},
}

DEFAULT_PROVIDER = 'ea520f49-36bc-4821-a16a-38bf0340d1f3'
DEFAULT_EXTRA_WORDS = 'album'

RE_VALIDATE_TITLE = re.compile(r'^[^\s"|][^"|]*[^\s"|]$')
RE_VALIDATE_URL = re.compile(r'^[^\s"]*%search%[^\s"]*$')

KEY_PROVIDER = 'search_engine_lookup_provider'
KEY_PROVIDERS = 'search_engine_lookup_providers'
KEY_EXTRA = 'search_engine_lookup_extra_words'


def show_popup(title='', content='', window=None):
    QtWidgets.QMessageBox.information(
        window,
        title,
        content,
        QtWidgets.QMessageBox.Ok,
        QtWidgets.QMessageBox.Ok
    )


def lookup_error():
    log.error("{0}: No existing metadata to lookup.".format(PLUGIN_NAME,))
    show_popup(_('Lookup Error'), _('There is no existing data to use for a search.'))


def do_lookup(text):
    provider = config.setting[KEY_PROVIDER]
    providers = config.setting[KEY_PROVIDERS]
    base_url = providers[provider]['url'] if provider in providers else DEFAULT_PROVIDERS[DEFAULT_PROVIDER]['url']
    url = base_url.replace(r'%search%', quote_plus(text))
    log.debug("{0}: Looking up {1}".format(PLUGIN_NAME, url,))
    _open(url)


def lookup_cover_art(title, artist):
    text = "{0} by {1} album cover".format(title, artist)
    do_lookup(text)


class SearchEngineLookupTest(BaseAction):
    NAME = 'Search engine lookup'

    def callback(self, cluster_list):
        extra = config.setting[KEY_EXTRA].split()
        for cluster in cluster_list:
            if isinstance(cluster, Cluster):
                parts = []
                if 'albumartist' in cluster.metadata and cluster.metadata['albumartist']:
                    parts.extend(cluster.metadata['albumartist'].split())
                if 'album' in cluster.metadata and cluster.metadata['albumartist']:
                    parts.extend(cluster.metadata['album'].split())
                if parts:
                    if extra:
                        parts.extend(extra)
                    text = ' '.join(parts)
                    do_lookup(text)
                else:
                    lookup_error()
            else:
                log.error("{0}: Argument is not a cluster. {1}".format(PLUGIN_NAME, cluster,))
                show_popup(_('Lookup Error'), _('There was a problem with the information provided for the cluster.'))


class AlbumCoverArtLookup(BaseAction):
    NAME = 'Album cover art lookup'

    def callback(self, album):
        metadata = album[0].metadata
        if 'album' in metadata and 'albumartist' in metadata:
            lookup_cover_art(metadata['album'], metadata['albumartist'])
        else:
            lookup_error()


class TrackCoverArtLookup(BaseAction):
    NAME = 'Track cover art lookup'

    def callback(self, track):
        metadata = track[0].metadata
        if 'title' in metadata and 'artist' in metadata:
            lookup_cover_art(metadata['title'], metadata['artist'])
        else:
            lookup_error()


class SearchEngineEditDialog(QtWidgets.QDialog):

    def __init__(self, parent=None, edit_provider='', edit_url='', titles=None):
        super().__init__(parent)
        self.parent = parent
        self.output = None
        self.edit_provider = edit_provider
        self.edit_url = edit_url
        self.providers = titles if titles else []

        self.valid_no = QtWidgets.QApplication.style().standardIcon(QtWidgets.QStyle.SP_DialogCancelButton).pixmap(16, 16)
        self.valid_yes = QtWidgets.QApplication.style().standardIcon(QtWidgets.QStyle.SP_DialogApplyButton).pixmap(16, 16)

        self.ui = Ui_SearchEngineEditorDialog()
        self.ui.setupUi(self)
        self.ui.le_title.setText(self.edit_provider)
        self.ui.le_url.setText(self.edit_url)

        self.setup_actions()
        self.check_validation()

    def setup_actions(self):
        self.ui.le_title.textChanged.connect(self.title_text_changed)
        self.ui.le_url.textChanged.connect(self.url_text_changed)
        self.ui.pb_save.clicked.connect(self.accept)
        self.ui.pb_cancel.clicked.connect(self.reject)

    def check_validation(self):
        valid_title = re.match(RE_VALIDATE_TITLE, self.edit_provider) and self.edit_provider not in self.providers
        self.ui.img_valid_title.setPixmap(self.valid_yes if valid_title else self.valid_no)

        valid_url = re.match(RE_VALIDATE_URL, self.edit_url)
        self.ui.img_valid_url.setPixmap(self.valid_yes if valid_url else self.valid_no)

        # Note that this needs to be forced to a bool to avoid Qt crashing Picard
        self.ui.pb_save.setEnabled(bool(valid_title and valid_url))

    def get_output(self):
        return self.output

    def accept(self):
        self.output = (self.edit_provider.strip(), self.edit_url.strip())
        super().accept()

    def title_text_changed(self, text):
        self.edit_provider = text
        self.check_validation()

    def url_text_changed(self, text):
        self.edit_url = text
        self.check_validation()


class SearchEngineLookupOptionsPage(OptionsPage):

    NAME = "search_engine_lookup_options"
    TITLE = "Search Engine Lookup"
    PARENT = "plugins"
    HELP_URL = "https://github.com/metabrainz/picard-plugins/blob/2.0/plugins/search_engine_lookup/README.md"

    options = [
        config.Option("setting", KEY_PROVIDERS, DEFAULT_PROVIDERS.copy()),
        config.TextOption("setting", KEY_PROVIDER, DEFAULT_PROVIDER),
        config.TextOption("setting", KEY_EXTRA, DEFAULT_EXTRA_WORDS),
    ]

    def __init__(self, parent=None):
        super(SearchEngineLookupOptionsPage, self).__init__(parent)
        self.ui = Ui_SearchEngineLookupOptionsPage()
        self.ui.setupUi(self)
        self.setup_actions()
        self.provider = ''
        self.providers = {}
        self.additional_words = ''

    def setup_actions(self):
        self.ui.list_providers.itemChanged.connect(self.select_provider)
        self.ui.pb_add.clicked.connect(self.add_provider)
        self.ui.pb_edit.clicked.connect(self.edit_provider)
        self.ui.pb_delete.clicked.connect(self.delete_provider)
        self.ui.pb_test.clicked.connect(self.test_provider)

    def load(self):
        # Settings for search engine providers
        self.providers = config.setting[KEY_PROVIDERS] or DEFAULT_PROVIDERS.copy()

        # Settings for search engine provider
        self.provider = config.setting[KEY_PROVIDER]
        if self.provider not in self.providers:
            # Assign an arbitrary valid value to self.provider
            self.provider = list(self.providers)[0]

        # Settings for search extra words
        self.additional_words = config.setting[KEY_EXTRA]
        self.ui.le_additional_words.setText(self.additional_words)

        # Display list of providers
        self.update_list()

    def select_provider(self, list_item):
        if list_item.checkState() == QtCore.Qt.Checked:
            # New provider selected
            self.provider = list_item.data(QtCore.Qt.UserRole)
            self.update_list(current_item=self.provider)
        else:
            # Attempt to deselect the current provider leaving none selected
            list_item.setCheckState(QtCore.Qt.Checked)

    def add_provider(self):
        provider_id = uuid4()
        self.edit_provider_dialog(provider_id)

    def edit_provider(self):
        current_item = self.ui.list_providers.currentItem()
        provider = current_item.text()
        provider_id = current_item.data(QtCore.Qt.UserRole)
        url = self.providers[provider_id]['url']
        self.edit_provider_dialog(provider_id, provider, url)

    def edit_provider_dialog(self, provider_id='', provider='', url=''):
        # List of titles currently used and not allowed.  Omit current title from the list when editing.
        titles = [x['name'] for x in self.providers.values() if x['name'] != provider]
        dialog = SearchEngineEditDialog(self, provider, url, titles)
        temp = dialog.exec_()
        if temp == QtWidgets.QDialog.Accepted:
            data = dialog.get_output()
            if data:
                new_provider, new_url = data
                self.providers[provider_id] = {'name': new_provider, 'url': new_url}
                self.update_list(provider_id)

    def delete_provider(self):
        current_item = self.ui.list_providers.currentItem()
        provider = current_item.text()
        provider_id = current_item.data(QtCore.Qt.UserRole)
        if current_item.checkState() or provider_id == self.provider:
            QtWidgets.QMessageBox.critical(
                self,
                _('Deletion Error'),
                _('You cannot delete the currently selected search provider.'),
                QtWidgets.QMessageBox.Ok,
                QtWidgets.QMessageBox.Ok
            )
        else:
            if QtWidgets.QMessageBox.warning(
                self,
                _('Confirm Deletion'),
                _('You are about to permanently delete the search provider "{provider_name}".  Continue?').format(provider_name=provider),
                QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel,
                QtWidgets.QMessageBox.Cancel
            ) == QtWidgets.QMessageBox.Ok:
                self.providers.pop(provider_id, None)
                self.update_list()

    def test_provider(self):
        current_item = self.ui.list_providers.currentItem()
        parts = ('The Beatles Abby Road ' + self.additional_words).strip().split()
        url = self.providers[current_item.data(QtCore.Qt.UserRole)]['url'].replace(r'%search%', quote_plus(' '.join(parts)))
        _open(url)

    def update_list(self, current_item=None):
        current_row = -1
        self.ui.list_providers.clear()
        for counter, provider_id in enumerate(self.providers):
            item = QtWidgets.QListWidgetItem()
            item.setFlags(item.flags() | QtCore.Qt.ItemIsUserCheckable)
            item.setText(self.providers[provider_id]['name'])
            item.setCheckState(QtCore.Qt.Checked if provider_id == self.provider else QtCore.Qt.Unchecked)
            item.setData(QtCore.Qt.UserRole, provider_id)
            self.ui.list_providers.addItem(item)
            if current_item and provider_id == current_item:
                current_row = counter
        current_row = max(current_row, 0)
        self.ui.list_providers.setCurrentRow(current_row)
        self.ui.list_providers.sortItems()

    def save(self):
        self._set_settings(config.setting)

    def _set_settings(self, settings):
        settings[KEY_PROVIDER] = self.provider.strip()
        settings[KEY_EXTRA] = self.additional_words.strip()
        settings[KEY_PROVIDERS] = self.providers or DEFAULT_PROVIDERS.copy()


register_cluster_action(SearchEngineLookupTest())
register_options_page(SearchEngineLookupOptionsPage)
register_album_action(AlbumCoverArtLookup())
register_track_action(TrackCoverArtLookup())



# -*- coding: utf-8 -*-

PLUGIN_NAME = 'View script variables'
PLUGIN_AUTHOR = 'Sophist'
PLUGIN_DESCRIPTION = '''Display a dialog box listing the metadata variables for the track / file.<br /><br />
This allows you to see metadata variables beginning with "~" which are not normally visible in the metadata
pane of the main Picard window, which can be useful when you are writing tagging or file naming scripts.
'''
PLUGIN_VERSION = '0.7.2'
PLUGIN_API_VERSIONS = ['2.0']
PLUGIN_LICENSE = "GPL-2.0"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.html"

from PyQt5 import QtWidgets, QtCore
try:
    from picard.util.tags import PRESERVED_TAGS
except ImportError:
    from picard.file import File
    PRESERVED_TAGS = File._default_preserved_tags

from picard.file import File
from picard.track import Track
from picard.ui.itemviews import BaseAction, register_file_action, register_track_action
from picard.plugins.viewvariables.ui_variables_dialog import Ui_VariablesDialog


class ViewVariables(BaseAction):
    NAME = 'View script variables'

    def callback(self, objs):
        obj = objs[0]
        files = self.tagger.get_files_from_objects(objs)
        if files:
            obj = files[0]
        dialog = ViewVariablesDialog(obj)
        dialog.exec_()


class ViewVariablesDialog(QtWidgets.QDialog):

    def __init__(self, obj, parent=None):
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_VariablesDialog()
        self.ui.setupUi(self)
        self.ui.buttonBox.accepted.connect(self.accept)
        self.ui.buttonBox.rejected.connect(self.reject)
        metadata = obj.metadata
        if isinstance(obj, File):
            self.setWindowTitle(_("File: %s") % obj.base_filename)
        elif isinstance(obj, Track):
            tn = metadata['tracknumber']
            if len(tn) == 1:
                tn = "0" + tn
            self.setWindowTitle(_("Track: %s %s ") % (tn, metadata['title']))
        else:
            self.setWindowTitle(_("Variables"))
        self._display_metadata(metadata)

    def _display_metadata(self, metadata):
        keys = metadata.keys()
        keys = sorted(keys, key=lambda key:
                      '0' + key if key in PRESERVED_TAGS and key.startswith('~') else
                      '1' + key if key.startswith('~') else
                      '2' + key)
        media = hidden = album = False
        table = self.ui.metadata_table
        key_example, value_example = self.get_table_items(table, 0)
        self.key_flags = key_example.flags()
        self.value_flags = value_example.flags()
        table.setRowCount(len(keys) + 3)
        i = 0
        for key in keys:
            if key in PRESERVED_TAGS and key.startswith('~'):
                if not media:
                    self.add_separator_row(table, i, _("File variables"))
                    i += 1
                    media = True
            elif key.startswith('~'):
                if not hidden:
                    self.add_separator_row(table, i, _("Hidden variables"))
                    i += 1
                    hidden = True
            else:
                if not album:
                    self.add_separator_row(table, i, _("Tag variables"))
                    i += 1
                    album = True

            key_item, value_item = self.get_table_items(table, i)
            i += 1
            key_item.setText("_" + key[1:] if key.startswith('~') else key)
            if key in metadata:
                value = metadata.getall(key)
                if len(value) == 1 and value[0] != '':
                    value = value[0]
                else:
                    value = repr(value)
                value_item.setText(value)

    def add_separator_row(self, table, i, title):
        key_item, value_item = self.get_table_items(table, i)
        font = key_item.font()
        font.setBold(True)
        key_item.setFont(font)
        key_item.setText(title)

    def get_table_items(self, table, i):
        key_item = table.item(i, 0)
        value_item = table.item(i, 1)
        if not key_item:
            key_item = QtWidgets.QTableWidgetItem()
            key_item.setFlags(self.key_flags)
            table.setItem(i, 0, key_item)
        if not value_item:
            value_item = QtWidgets.QTableWidgetItem()
            value_item.setFlags(self.value_flags)
            table.setItem(i, 1, value_item)
        return key_item, value_item

vv = ViewVariables()
register_file_action(vv)
register_track_action(vv)



PLUGIN_NAME = "Padded disc and tracknumbers"
PLUGIN_AUTHOR = "Wieland Hoffmann"
PLUGIN_DESCRIPTION = """
Adds padded disc- and tracknumbers so the length of all disc- and tracknumbers
is the same. They are stored in the `_paddedtracknumber` and `_paddeddiscnumber`
tags."""

PLUGIN_VERSION = "1.0.1"
PLUGIN_API_VERSIONS = ["0.15.0", "0.15.1", "0.16.0", "1.0.0", "1.1.0", "1.2.0",
                       "1.3.0", "2.0", ]
PLUGIN_LICENSE = "GPL-2.0-or-later"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.html"

from picard.metadata import register_track_metadata_processor


@register_track_metadata_processor
def add_padded_tn(album, metadata, track, release):
    maxlength = len(metadata["totaltracks"])
    islength = len(metadata["tracknumber"])
    metadata["~paddedtracknumber"] = (int(maxlength - islength) * "0" +
                                      metadata["tracknumber"])


@register_track_metadata_processor
def add_padded_dn(album, metadata, track, release):
    maxlength = len(metadata["totaldiscs"])
    islength = len(metadata["discnumber"])
    metadata["~paddeddiscnumber"] = (int(maxlength - islength) * "0" +
                                     metadata["discnumber"])
